import typing


class Counter:
    def __init__(self) -> None:
        self._count = 0

    def get_count(self, *args: typing.Any, **kwargs: typing.Any) -> int:
        """
        Return the number of times this method was called.

        Returns:
            The number of times the method was called.
        """
        self._count += 1
        return self._count



"""
Plotters
========

#. :class:`.ProgressPlotter`
#. :class:`.SelectionPlotter`


The are multiple different plotters, but they are not related by an
inheritance hierarchy. Each plotter has its own distinct API, according
to its needs.

"""



import itertools as it


def test_get_core_atom_ids(case_data):
    """
    Test :meth:`.FunctionalGroup.get_core_atom_ids`.

    Parameters
    ----------
    case_data : :class:`.CaseData`
        The test case. Holds the functional group to test and the
        atoms holding the correct ids.

    Returns
    -------
    None : :class:`NoneType`

    """

    _test_get_core_atom_ids(
        functional_group=case_data.functional_group,
        core_atoms=case_data.core_atoms,
    )


def _test_get_core_atom_ids(functional_group, core_atoms):
    """
    Test :meth:`.FunctionalGroup.get_core_atom_ids`.

    Parameters
    ----------
    functional_group : :class:`.FunctionalGroup`
        The functional group to test.

    core_atoms : :class:`tuple` of :class:`.Atom`
        Atoms with the correct ids.

    Returns
    -------
    None : :class:`NoneType`

    """

    for id_, core_atom in it.zip_longest(
        functional_group.get_core_atom_ids(),
        core_atoms,
    ):
        assert id_ == core_atom.get_id()



import itertools as it

import pytest
import stk

from ..case_data import CaseData


@pytest.fixture
def dative_reaction(
    functional_group1,
    functional_group1_2,
    bond_order,
    periodicity,
):
    """
    A :class:`.CaseData` instance.

    """

    return CaseData(
        reaction=stk.DativeReaction(
            reaction=stk.OneOneReaction(
                functional_group1=functional_group1,
                functional_group2=functional_group1_2,
                bond_order=bond_order,
                periodicity=periodicity,
            ),
        ),
        new_atoms=(),
        new_bonds=(
            get_bond(
                functional_group1=functional_group1,
                functional_group2=functional_group1_2,
                bond_order=bond_order,
                periodicity=periodicity,
            ),
        ),
        deleted_atoms=tuple(
            it.chain(
                functional_group1.get_deleters(),
                functional_group1_2.get_deleters(),
            )
        ),
        deleted_bonds=(),
    )


def get_bond(
    functional_group1,
    functional_group2,
    bond_order,
    periodicity,
):
    (bonder1,) = functional_group1.get_bonders()
    (bonder2,) = functional_group2.get_bonders()

    if is_metal(bonder1):
        return stk.Bond(
            atom1=bonder2,
            atom2=bonder1,
            order=bond_order,
            periodicity=periodicity,
        )
    else:
        return stk.Bond(
            atom1=bonder1,
            atom2=bonder2,
            order=bond_order,
            periodicity=periodicity,
        )


def is_metal(atom):
    metal_atomic_numbers = set(
        it.chain(
            range(21, 31),
            range(39, 49),
            range(72, 81),
        )
    )
    return atom.get_atomic_number() in metal_atomic_numbers



import pytest
import stk

from ..case_data import CaseData


class CofData:
    """
    Data to initialize a COF :class:`.CaseData` instance.

    Attributes
    ----------
    topology_graph : :class:`type`
        A COF class.

    building_blocks : :class:`tuple` of :class:`.BuildingBlock`
        The building blocks of the COF.

    lattice_size : :class:`tuple` of :class:`int`
        The size of the lattice.

    vertex_alignments : :class:`dict`
        Passed to the `vertex_alignments` parameter of the COF
        initializer.

    num_new_atoms : :class:`int`
        The number of new atoms added by the construction process.

    num_new_bonds : :class:`int`
        The number of new bonds added by the construction process.

    num_building_blocks : :class:`dict`
        For each building block in :attr:`building_blocks`, maps its
        index to the number of times its used in the construction of
        the COF.

    """

    def __init__(
        self,
        topology_graph,
        building_blocks,
        lattice_size,
        vertex_alignments,
        num_new_atoms,
        num_new_bonds,
        num_building_blocks,
    ):
        """
        Initialize a :class:`.CofData` instance.

        Parameters
        ----------
        topology_graph : :class:`type`
            A COF class.

        building_blocks : :class:`tuple` of :class:`.BuildingBlock`
            The building blocks of the COF.

        lattice_size : :class:`tuple` of :class:`int`
            The size of the lattice.

        vertex_alignments : :class:`dict`
            Passed to the `vertex_alignments` parameter of the COF
            initializer.

        num_new_atoms : :class:`int`
            The number of new atoms added by the construction process.

        num_new_bonds : :class:`int`
            The number of new bonds added by the construction process.

        num_building_blocks : :class:`dict`
            For each building block in `building_blocks`, maps its
            index to the number of times its used in the construction
            of the COF.

        """

        self.constructed_molecule = stk.ConstructedMolecule(
            topology_graph=topology_graph(
                building_blocks=building_blocks,
                lattice_size=lattice_size,
                vertex_alignments=vertex_alignments,
            )
        )
        self.num_new_atoms = num_new_atoms
        self.num_new_bonds = num_new_bonds
        self.num_building_blocks = {
            building_blocks[index]: num
            for index, num in num_building_blocks.items()
        }
        self.building_blocks = building_blocks

    @classmethod
    def init_from_construction_result(
        cls,
        topology_graph,
        building_blocks,
        lattice_size,
        vertex_alignments,
        num_new_atoms,
        num_new_bonds,
        num_building_blocks,
    ):
        """
        Initialize a :class:`.CofData` instance.

        This method creates the constructed molecule using
        :meth:`.ConstructedMolecule.init_from_construction_result`.

        Parameters
        ----------
        topology_graph : :class:`type`
            A COF class.

        building_blocks : :class:`tuple` of :class:`.BuildingBlock`
            The building blocks of the COF.

        lattice_size : :class:`tuple` of :class:`int`
            The size of the lattice.

        vertex_alignments : :class:`dict`
            Passed to the `vertex_alignments` parameter of the COF
            initializer.

        num_new_atoms : :class:`int`
            The number of new atoms added by the construction process.

        num_new_bonds : :class:`int`
            The number of new bonds added by the construction process.

        num_building_blocks : :class:`dict`
            For each building block in `building_blocks`, maps its
            index to the number of times its used in the construction
            of the COF.

        """

        obj = cls.__new__(cls)
        topology_graph_instance = topology_graph(
            building_blocks=building_blocks,
            lattice_size=lattice_size,
            vertex_alignments=vertex_alignments,
        )
        construction_result = topology_graph_instance.construct()
        obj.constructed_molecule = (
            stk.ConstructedMolecule.init_from_construction_result(
                construction_result=construction_result,
            )
        )
        obj.num_new_atoms = num_new_atoms
        obj.num_new_bonds = num_new_bonds
        obj.num_building_blocks = {
            building_blocks[index]: num
            for index, num in num_building_blocks.items()
        }
        obj.building_blocks = building_blocks
        return obj


@pytest.fixture(
    scope="session",
    params=(
        lambda: CofData(
            topology_graph=stk.cof.Honeycomb,
            building_blocks=(
                stk.BuildingBlock(
                    smiles=("Br[C+]1[C+2][C+](Br)[C+](F)[C+](Br)[C+2]1"),
                    functional_groups=[stk.BromoFactory()],
                ),
                stk.BuildingBlock(
                    smiles="Br[C+]=NC#CBr",
                    functional_groups=[stk.BromoFactory()],
                ),
            ),
            lattice_size=(2, 2, 1),
            vertex_alignments=None,
            num_new_atoms=0,
            num_new_bonds=20,
            num_building_blocks={0: 8, 1: 12},
        ),
        lambda: CofData.init_from_construction_result(
            topology_graph=stk.cof.Honeycomb,
            building_blocks=(
                stk.BuildingBlock(
                    smiles=("Br[C+]1[C+2][C+](Br)[C+](F)[C+](Br)[C+2]1"),
                    functional_groups=[stk.BromoFactory()],
                ),
                stk.BuildingBlock(
                    smiles="Br[C+]=NC#CBr",
                    functional_groups=[stk.BromoFactory()],
                ),
            ),
            lattice_size=(2, 2, 1),
            vertex_alignments=None,
            num_new_atoms=0,
            num_new_bonds=20,
            num_building_blocks={0: 8, 1: 12},
        ),
        lambda: CofData(
            topology_graph=stk.cof.Honeycomb,
            building_blocks=(
                stk.BuildingBlock(
                    smiles=("Br[C+]1[C+2][C+](Br)[C+](F)[C+](Br)[C+2]1"),
                    functional_groups=[stk.BromoFactory()],
                ),
                stk.BuildingBlock(
                    smiles="Br[C+]=NC#CBr",
                    functional_groups=[stk.BromoFactory()],
                ),
            ),
            lattice_size=(2, 2, 1),
            vertex_alignments={0: 1, 1: 1, 2: 1, 3: 1, 4: 1},
            num_new_atoms=0,
            num_new_bonds=20,
            num_building_blocks={0: 8, 1: 12},
        ),
        lambda: CofData(
            topology_graph=stk.cof.Honeycomb,
            building_blocks=(
                stk.BuildingBlock(
                    smiles=("Br[C+]1[C+2][C+](Br)[C+](F)[C+](Br)[C+2]1"),
                    functional_groups=[stk.BromoFactory()],
                ),
                stk.BuildingBlock(
                    smiles="Br[C+]=NC#CBr",
                    functional_groups=[stk.BromoFactory()],
                ),
            ),
            lattice_size=(2, 2, 1),
            vertex_alignments={0: 2, 1: 2},
            num_new_atoms=0,
            num_new_bonds=20,
            num_building_blocks={0: 8, 1: 12},
        ),
        lambda: CofData(
            topology_graph=stk.cof.LinkerlessHoneycomb,
            building_blocks=(
                stk.BuildingBlock(
                    smiles=("Br[C+]1[C+2][C+](Br)[C+](F)[C+](Br)[C+2]1"),
                    functional_groups=[stk.BromoFactory()],
                ),
            ),
            lattice_size=(2, 2, 1),
            vertex_alignments=None,
            num_new_atoms=0,
            num_new_bonds=8,
            num_building_blocks={0: 8},
        ),
        lambda: CofData(
            topology_graph=stk.cof.Hexagonal,
            building_blocks=(
                stk.BuildingBlock(
                    smiles=("Br[C+]1[C+](F)[C+](I)[C+](I)[C+](Br)C1Br"),
                    functional_groups=[
                        stk.BromoFactory(),
                        stk.IodoFactory(),
                        stk.FluoroFactory(),
                    ],
                ),
                stk.BuildingBlock(
                    smiles="Br[C+]=NC#CBr",
                    functional_groups=[stk.BromoFactory()],
                ),
            ),
            lattice_size=(2, 2, 1),
            vertex_alignments={0: 5},
            num_new_atoms=0,
            num_new_bonds=81,
            num_building_blocks={0: 16, 1: 48},
        ),
        lambda: CofData(
            topology_graph=stk.cof.Kagome,
            building_blocks=(
                stk.BuildingBlock(
                    smiles=("Br[C+]1[C+](Br)[C+](F)[C+](Br)[C+](Br)" "[C+2]1"),
                    functional_groups=[stk.BromoFactory()],
                ),
                stk.BuildingBlock(
                    smiles="Br[C+]=NC#CBr",
                    functional_groups=[stk.BromoFactory()],
                ),
            ),
            lattice_size=(2, 2, 1),
            vertex_alignments=None,
            num_new_atoms=0,
            num_new_bonds=41,
            num_building_blocks={0: 12, 1: 24},
        ),
        lambda: CofData(
            topology_graph=stk.cof.Square,
            building_blocks=(
                stk.BuildingBlock(
                    smiles="BrC1=C(Br)C(F)(Br)[C+]1Br",
                    functional_groups=[stk.BromoFactory()],
                ),
                stk.BuildingBlock(
                    smiles="Br[C+]=NC#CBr",
                    functional_groups=[stk.BromoFactory()],
                ),
            ),
            lattice_size=(2, 2, 1),
            vertex_alignments=None,
            num_new_atoms=0,
            num_new_bonds=12,
            num_building_blocks={0: 4, 1: 8},
        ),
    ),
)
def cof_data(request) -> CofData:
    """
    A :class:`.CofData` instance.

    """

    return request.param()


@pytest.fixture
def cof(cof_data):
    """
    A :class:`.CaseData` instance.

    """

    return CaseData(
        constructed_molecule=cof_data.constructed_molecule,
        num_new_atoms=cof_data.num_new_atoms,
        num_new_bonds=cof_data.num_new_bonds,
        num_building_blocks=cof_data.num_building_blocks,
        building_blocks=cof_data.building_blocks,
    )



import typing

from stk._internal.constructed_molecule import ConstructedMolecule
from stk._internal.topology_graphs.topology_graph.topology_graph import (
    TopologyGraph,
)

T = typing.TypeVar("T", bound=TopologyGraph)


class MoleculeRecord(typing.Generic[T]):
    """
    An abstract base class for molecular records used by the EA.

    Notes:

        You might notice that the public methods of this abstract base
        class are implemented. This is a default implementation provided
        purely for convenience. Subclasses can freely ignore or
        override this implementation.
    """

    def __init__(self, topology_graph: T) -> None:
        """
        Parameters:
            topology_graph:
                The topology graph of a :class:`.ConstructedMolecule`.
        """
        self._molecule = ConstructedMolecule(topology_graph)
        self._topology_graph = topology_graph

    def get_molecule(self) -> ConstructedMolecule:
        """
        Get the molecule held by the record.

        Returns:
            The molecule held by the record.
        """
        return self._molecule

    def get_topology_graph(self) -> T:
        """
        Get the topology graph of the molecule.

        Returns:
            The topology graph.
        """
        return self._topology_graph

    def __repr__(self) -> str:
        return f"MoleculeRecord({self._topology_graph!r})"



import pytest
import stk

from ...case_data import CaseData


@pytest.fixture(
    scope="session",
    params=(
        lambda name: CaseData(
            molecule=stk.ConstructedMolecule(
                topology_graph=stk.cof.LinkerlessHoneycomb(
                    building_blocks=(
                        stk.BuildingBlock(
                            smiles=(
                                "Br[C+]1C2[C+]=NC2(Br)[C+](F)[C+](Br)" "[C+2]1"
                            ),
                            functional_groups=[stk.BromoFactory()],
                        ),
                    ),
                    lattice_size=(2, 2, 1),
                ),
            ),
            smiles=(
                "[H]C12[C+]=NC1(Br)[C+](F)[C+](Br)[C+2][C+]2[C+]1[C+2]"
                "[C+]2[C+]3[C+2][C+]([C+]4[C+2][C+](Br)[C+](F)C5(N=[C+"
                "]C45[H])C45N=[C+]C4([H])[C+]([C+]4[C+2][C+](Br)[C+](F"
                ")C6(Br)N=[C+]C46[H])[C+2][C+]([C+]4[C+2][C+]([C+]6[C+"
                "2][C+](Br)[C+](F)C7(N=[C+]C67[H])C6(N=[C+]C16[H])[C+]"
                "2F)C1([H])[C+]=NC1(Br)[C+]4F)[C+]5F)C1([H])[C+]=NC1(B"
                "r)[C+]3F"
            ),
            name=name,
        ),
    ),
)
def cof_linkerless_honeycomb(request) -> CaseData:
    return request.param(
        f"{request.fixturename}{request.param_index}",
    )



"""
SMARTS Functional Group Factory
===============================

"""


from stk._internal.functional_group_factories.functional_group_factory import (
    FunctionalGroupFactory,
)
from stk._internal.functional_group_factories.utilities import get_atom_ids
from stk._internal.functional_groups.generic_functional_group import (
    GenericFunctionalGroup,
)


class SmartsFunctionalGroupFactory(FunctionalGroupFactory):
    """
    Creates :class:`.GenericFunctionalGroup` instances.

    Examples
    --------
    *Using SMARTS to Define Functional Groups*

    You want to create a building block which has
    :class:`.GenericFunctionalGroup` functional groups based on the
    SMARTS string: ``[Br][C]``.
    You want the ``C`` atom to be the *bonder* atom, and the
    ``Br`` atom to be the *deleter* atom.

    .. testcode:: using-smarts-to-define-functional-groups

        import stk

        building_block = stk.BuildingBlock(
            smiles='BrCCCBr',
            functional_groups=(
                stk.SmartsFunctionalGroupFactory(
                    smarts='[Br][C]',
                    bonders=(1, ),
                    deleters=(0, ),
                ),
            ),
        )

    .. testcode:: using-smarts-to-define-functional-groups
        :hide:

        fg1, fg2 = building_block.get_functional_groups()
        assert fg1.get_num_bonders() == 1
        assert sum(1 for _ in fg1.get_deleters()) == 1
        assert fg2.get_num_bonders() == 1
        assert sum(1 for _ in fg2.get_deleters()) == 1

        assert all(
            isinstance(atom, stk.C)
            for functional_group
            in building_block.get_functional_groups()
            for atom
            in functional_group.get_bonders()
        )
        assert all(
            isinstance(atom, stk.Br)
            for functional_group
            in building_block.get_functional_groups()
            for atom
            in functional_group.get_deleters()
        )

    See Also
    --------
    :class:`.GenericFunctionalGroup`
        Defines *bonders* and  *deleters*.

    """

    def __init__(self, smarts, bonders, deleters, placers=None):
        """
        Initialize a :class:`.SmartsFunctionalGroupFactory` instance.

        Parameters
        ----------
        smarts : :class:`str`
            The SMARTS defining the functional group.

        bonders : :class:`tuple` of :class:`int`
            The indices of atoms in `smarts`, which are *bonder* atoms.

        deleters : :class:`tuple` of :class:`int`
            The indices of atoms in `smarts`, which are *deleter*
            atoms.

        placers : :class:`tuple` of :class:`int`, optional
            The indices of atoms in `smarts`, which are *placer* atoms.
            If ``None``, the *bonder* atoms will be used.

        """

        self._smarts = smarts
        self._bonders = bonders
        self._deleters = deleters
        self._placers = bonders if placers is None else placers

    def get_functional_groups(self, molecule):
        for atom_ids in get_atom_ids(self._smarts, molecule):
            atoms = tuple(molecule.get_atoms(atom_ids))
            yield GenericFunctionalGroup(
                atoms=atoms,
                bonders=tuple(atoms[i] for i in self._bonders),
                deleters=tuple(atoms[i] for i in self._deleters),
                placers=tuple(atoms[i] for i in self._placers),
            )



"""
Graph State
===========

"""

from collections import defaultdict

import numpy as np


class GraphState:
    """
    The topology graph of a molecule under construction.

    """

    __slots__ = [
        "_vertex_building_blocks",
        "_vertices",
        "_edges",
        "_lattice_constants",
        "_vertex_edges",
        "_num_building_blocks",
    ]

    def __init__(
        self,
        building_block_vertices,
        edges,
        lattice_constants,
    ):
        """
        Initialize a :class:`._GraphState` instance.

        Parameters
        ----------
        building_block_vertices : :class:`dict`
            Maps each :class:`.BuildingBlock` to be placed, to a
            :class:`tuple` of :class:`.Vertex` instances, on which
            it should be placed.

        edges : :class:`tuple` of :class:`.Edge`
            The edges which make up the topology graph.

        lattice_constants : :class:`tuple` of :class:`numpy.ndarray`
            A :class:`numpy.ndarray` for each lattice constant.
            Can be an empty :class:`tuple` if the topology graph is
            not periodic.

        """

        self._vertex_building_blocks = {
            vertex.get_id(): building_block
            for building_block, vertices in building_block_vertices.items()
            for vertex in vertices
        }
        self._num_building_blocks = {
            building_block: len(vertices)
            for building_block, vertices in building_block_vertices.items()
        }
        self._vertices = {
            vertex.get_id(): vertex
            for vertices in building_block_vertices.values()
            for vertex in vertices
        }
        self._edges = edges
        self._lattice_constants = tuple(
            map(
                np.array,
                lattice_constants,
            )
        )
        self._vertex_edges = self._get_vertex_edges()

    def _get_vertex_edges(self):
        """
        Get the edges connected to each vertex.

        Returns
        -------
        :class:`dict`
            Maps the id of every vertex to a :class:`tuple` of
            :class:`.Edge` instances connected to it.

        """

        vertex_edges = defaultdict(list)
        for edge in self._edges:
            if edge.is_periodic():
                for vertex_id in edge.get_vertex_ids():
                    periodic_edge = self._get_periodic_edge(
                        edge=edge,
                        reference=vertex_id,
                    )
                    vertex_edges[vertex_id].append(periodic_edge)
            else:
                for vertex_id in edge.get_vertex_ids():
                    vertex_edges[vertex_id].append(edge)
        return vertex_edges

    def _get_periodic_edge(self, edge, reference):
        """
        Get an :class:`.Edge`, with its position correctly set.

        For a periodic edge, its correct position is not at the
        midpoint of the two vertices it connects. Instead, its
        correction position is different for each vertex. The get the
        correct position from the perspective of *vertex1*, *vertex2*
        must first be shifted to its periodic position, and only then
        can the midpoint of the vertices be used to get the edge
        position. An analogous calculation must be done to get the
        position of the edge from the perspective of *vertex2*.

        Parameters
        ----------
        edge : :class:`.Edge`
            The edge whose periodic position must be set.

        reference : :class:`int`
            The id of the vertex, relative to which the edge position
            is being calculated.

        Returns
        -------
        :class:`.Edge`
            A clone of `edge`, shifted to have the correct periodic
            position relative to `reference`.

        """

        vertex1 = self._vertices[reference]
        id1, id2 = edge.get_vertex_ids()
        vertex2 = self._vertices[id1 if reference == id2 else id2]

        direction = 1 if reference == id1 else -1
        periodicity = np.array(edge.get_periodicity())
        end_cell = vertex1.get_cell() + direction * periodicity
        cell_shift = end_cell - vertex2.get_cell()

        shift = sum(
            axis_shift * constant
            for axis_shift, constant in zip(
                cell_shift,
                self._lattice_constants,
            )
        )
        position = (
            vertex2.get_position() + shift + vertex1.get_position()
        ) / 2
        return edge.with_position(position)

    def clone(self):
        """
        Get a clone.

        Returns
        -------
        :class:`._GraphState`
            The clone. Has the same type as the original instance.

        """

        clone = self.__class__.__new__(self.__class__)
        clone._vertex_building_blocks = dict(self._vertex_building_blocks)
        clone._vertices = dict(self._vertices)
        clone._vertex_edges = dict(self._vertex_edges)
        clone._edges = self._edges
        clone._lattice_constants = tuple(
            map(
                np.array,
                self._lattice_constants,
            )
        )
        clone._num_building_blocks = dict(self._num_building_blocks)
        return clone

    def get_building_block(self, vertex_id):
        """
        Get the building block to be placed on a given vertex.

        Parameters
        ----------
        vertex_id : :class:`int`
            The id of the vertex, on which the building block is to
            be placed.

        Returns
        -------
        :class:`.BuildingBlock`
            The building block.

        """

        return self._vertex_building_blocks[vertex_id]

    def get_vertices(self, vertex_ids=None):
        """
        Yield the topology graph vertices.

        Parameters
        ----------
        vertex_ids : :class:`iterable` of :class:`int`, optional
            The ids of vertices to yield. If ``None``, all vertices
            will be yielded. Can be a single :class:`int` if a
            single vertex is to be yielded.

        Yields
        ------
        :class:`.Vertex`
            A vertex.

        """

        if vertex_ids is None:
            vertex_ids = range(len(self._vertices))
        elif isinstance(vertex_ids, int):
            vertex_ids = (vertex_ids,)

        for vertex_id in vertex_ids:
            yield self._vertices[vertex_id]

    def get_num_vertices(self):
        """
        Get the number of vertices in the topology graph.

        Returns
        -------
        :class:`int`
            The number of vertices in the topology graph.

        """

        return len(self._vertices)

    def get_edge(self, edge_id):
        """
        Get an edge.

        Parameters
        ----------
        edge_id : :class:`int`
            The id of an edge.

        Returns
        -------
        :class:`.Edge`
            An edge.

        """

        return self._edges[edge_id]

    def get_num_edges(self):
        """
        Get the number of edges in the topology graph.

        Returns
        -------
        :class:`int`
            The number of edges.

        """

        return len(self._edges)

    def get_lattice_constants(self):
        """
        Get the lattice constants of the state.

        Returns
        -------
        :class:`tuple` of :class:`numpy.ndarray`
            The lattice constants.

        """

        return tuple(map(np.array, self._lattice_constants))

    def get_edges(self, vertex_id):
        """
        Get the edges connect to a vertex.

        Parameters
        ----------
        vertex_id : class:`int`
            The id of a vertex.

        Returns
        -------
        :class:`tuple` of :class:`.Edge`
            The connected edges.

        """

        return self._vertex_edges[vertex_id]

    def _with_vertices(self, vertices):
        """
        Modify the instance.

        """

        self._vertices = {vertex.get_id(): vertex for vertex in vertices}
        return self

    def with_vertices(self, vertices):
        """
        Returns a clone holding `vertices`.

        Parameters
        ----------
        vertices : :class:`iterable` of :class:`.Vertex`
            The vertices the clone should hold.

        Returns
        -------
        :class:`._GraphState`
            The clone. Has the same type as the original instance.

        """

        return self.clone()._with_vertices(vertices)

    def _with_lattice_constants(self, lattice_constants):
        """
        Modify the instance.

        """

        self._lattice_constants = tuple(
            map(
                np.array,
                lattice_constants,
            )
        )
        return self

    def with_lattice_constants(self, lattice_constants):
        """
        Return a clone holding the `lattice_constants`.

        Parameters
        ----------
        lattice_constants : :class:`tuple` of :class:`numpy.ndarray`
            The lattice constants of the clone. Requires 3 arrays of
            size``(3, )``.

        Returns
        -------
        :class:`._GraphState`
            The clone holding the new lattice constants. Has the same
            type as the original instance.

        """

        return self.clone()._with_lattice_constants(lattice_constants)

    def get_num_building_block(self, building_block):
        """
        Get the number of times `building_block` is present.

        Parameters
        ----------
        building_block : :class:`.BuildingBlock`
            The building block whose frequency in the topology graph
            is desired.

        Returns
        -------
        :class:`int`
            The number of times `building_block` is present in the
            topology graph.

        """

        return self._num_building_blocks[building_block]

    def get_building_blocks(self):
        """
        Yield the building blocks.

        Building blocks are yielded in an order based on their
        position in the topology graph. For two equivalent
        topology graphs, but with different building blocks,
        equivalently positioned building blocks will be yielded at the
        same time.

        Yields
        ------
        :class:`.BuildingBlock`
            A building block of the topology graph.

        """

        yielded = set()
        for vertex_id in range(max(self._vertex_building_blocks) + 1):
            building_block = self._vertex_building_blocks[vertex_id]
            if building_block not in yielded:
                yielded.add(building_block)
                yield building_block



from dataclasses import dataclass
from typing import Any

import stk


@dataclass(frozen=True, slots=True)
class CaseData:
    fitness_normalizer: stk.FitnessNormalizer[stk.MoleculeRecord[Any]]
    fitness_values: dict[stk.MoleculeRecord[Any], Any]
    normalized: dict[stk.MoleculeRecord[Any], Any]

    @staticmethod
    def new(
        fitness_normalizer: stk.FitnessNormalizer[stk.MoleculeRecord[Any]],
        fitness_values: dict[stk.MoleculeRecord[Any], tuple[Any, Any]],
    ) -> "CaseData":
        return CaseData(
            fitness_normalizer=fitness_normalizer,
            fitness_values={
                record: fitness_value[0]
                for record, fitness_value in fitness_values.items()
            },
            normalized={
                record: fitness_value[1]
                for record, fitness_value in fitness_values.items()
            },
        )



import pytest
import stk

from ..case_data import GenericCaseData


@pytest.fixture
def aldehyde(get_atom_ids):
    a, b, c, d = get_atom_ids(4)
    return _aldehyde(stk.C(a), stk.O(b), stk.H(c), stk.C(d))


def _aldehyde(carbon, oxygen, hydrogen, atom):
    bonders = (carbon,)
    deleters = (oxygen,)
    return GenericCaseData(
        functional_group=stk.Aldehyde(
            carbon=carbon,
            oxygen=oxygen,
            hydrogen=hydrogen,
            atom=atom,
            bonders=bonders,
            deleters=deleters,
        ),
        atoms=(carbon, oxygen, hydrogen, atom),
        bonders=bonders,
        deleters=deleters,
    )



import pytest


@pytest.fixture(
    params=(
        lambda molecule: None,
        lambda molecule: range(molecule.get_num_atoms()),
        lambda molecule: range(0, molecule.get_num_atoms(), 2),
        lambda molecule: list(range(0, min(1, molecule.get_num_atoms()))),
        lambda molecule: tuple(range(0, min(1, molecule.get_num_atoms()))),
        lambda molecule: (
            i for i in range(0, min(1, molecule.get_num_atoms()))
        ),
        pytest.param(
            lambda molecule: (),
            marks=pytest.mark.xfail(strict=True, raises=ValueError),
        ),
        lambda molecule: range(min(molecule.get_num_atoms(), 1)),
    ),
)
def get_atom_ids(request):
    """
    Return an atom_ids parameter for a :class:`.Molecule`.

    Parameters
    ----------
    molecule : :class:`.Molecule`
        The molecule for which `atom_ids` are returned.

    Returns
    -------
    :class:`iterable` of :class:`int`
        An `atom_ids` parameter.

    """

    return request.param


@pytest.fixture
def centroid(origin):
    return origin



import stk


class MockGraphState(stk.GraphState):
    def __init__(self, edges):
        self._edges = edges


class MockConstructionState(stk.ConstructionState):
    def __init__(
        self,
        edges,
        edge_functional_groups,
        position_matrix=None,
    ):
        self._molecule_state = stk.MoleculeState()
        self._molecule_state._position_matrix = position_matrix
        self._molecule_state._edge_functional_groups = edge_functional_groups
        self._graph_state = MockGraphState(edges)


class MockEdge(stk.Edge):
    def __init__(self, id, periodicity):
        self._id = id
        self._periodicity = periodicity



from stk._internal.key_makers.molecule import MoleculeKeyMaker
from stk._internal.key_makers.utilities import get_inchi
from stk._internal.molecule import Molecule


class Inchi(MoleculeKeyMaker):
    """
    Used to get the InChI of molecules.

    Examples:

        *Adding InChI to a Molecule's JSON*

        You want to use the InChI as part of a JSON representation of a
        molecule

        .. testcode:: adding-inchi-to-a-molecules-json

            import stk

            jsonizer = stk.MoleculeJsonizer(
                key_makers=(stk.Inchi(), ),
            )
            building_block = stk.BuildingBlock('NCCN')
            # Get the JSON representation, including an InChI.
            json = jsonizer.to_json(building_block)

        .. testcode:: adding-inchi-to-a-molecules-json
            :hide:

            _inchi = stk.Inchi()
            _expected_inchi = 'InChI=1S/C2H8N2/c3-1-2-4/h1-4H2'
            _molecule_json = json['molecule']
            assert (
                _molecule_json[_inchi.get_key_name()]
                == _expected_inchi
            )

            _matrix_json = json['matrix']
            assert (
                _matrix_json[_inchi.get_key_name()]
                == _expected_inchi
            )

    """

    def __init__(self) -> None:
        return

    def get_key_name(self) -> str:
        return "InChI"

    def get_key(self, molecule: Molecule) -> str:
        return get_inchi(molecule)

    def __str__(self) -> str:
        return repr(self)

    def __repr__(self) -> str:
        return "Inchi()"



def test_get_canonical_atom_ids(case_data):
    """
    Test :meth:`.Molecule.get_canonical_atom_ids`.

    Parameters
    ----------
    case_data : :class:`.CaseData`
        A test case. Holds the molecule to test and the expected
        canonical atom ids.

    Returns
    -------
    None : :class:`NoneType`

    """

    _test_get_canonical_atom_ordering(
        molecule=case_data.molecule,
        canonical_atom_ids=case_data.canonical_atom_ids,
    )


def _test_get_canonical_atom_ordering(molecule, canonical_atom_ids):
    assert molecule.get_canonical_atom_ids() == canonical_atom_ids



from ..utilities import is_clone_functional_group


def test_clone(functional_group):
    """
    Test :meth:`.FunctionalGroup.clone`.

    Parameters
    ----------
    functional_group : :class:`.FunctionalGroup`
        The functional group to clone.

    Returns
    -------
    None : :class:`NoneType`

    """

    clone = functional_group.clone()
    is_clone_functional_group(functional_group, clone)



import numpy as np

from ....utilities import is_clone


def test_with_structure_from_file(datadir, case_data):
    """
    Test :meth:`.Molecule.with_structure_from_file`.

    Parameters
    ----------
    datadir : :class:`pathlib.Path`
        Holds the data files for the test.

    case_data : :class:`.CaseData`
        A test case. Holds the molecule to test and the path of its
        structure file.

    Returns
    -------
    None : :class:`NoneType`

    """

    # Save a copy of the position matrix, to ensure the original
    # molecule is not modified by the test, because it is meant to be
    # immutable.
    position_matrix = case_data.molecule.get_position_matrix()
    _test_with_structure_from_file(
        molecule=case_data.molecule,
        path=str(datadir / case_data.path),
    )
    assert np.all(
        np.equal(
            position_matrix,
            case_data.molecule.get_position_matrix(),
        )
    )


def _test_with_structure_from_file(molecule, path):
    """
    Test :meth:`.Molecule.with_structure_from_file`.

    Parameters
    ----------
    molecule : :class:`.Molecule`
        The molecule to test.

    path : :class:`str`
        The path the structure file of `molecule`.

    Returns
    -------
    None : :class:`NoneType`

    """

    new = molecule.with_structure_from_file(path)
    is_clone(new, molecule)
    size_diff = abs(
        molecule.get_maximum_diameter() - new.get_maximum_diameter()
    )
    assert size_diff > 1



import typing

T = typing.TypeVar("T")


class CrossoverRecord(typing.Generic[T]):
    """
    Abstract base class for a record of a crossover operation.

    Notes:

        You might notice that the public methods of this abstract base
        class are implemented. This is just a default implementation, which
        can be used directly by users and subclasses, but can also be
        freely replaced during subclass implementation, if need be.

    """

    def __init__(self, molecule_record: T, crosser_name: str) -> None:
        """
        Parameters:

            molecule_record:
                The molecule produced by the crossover operation.

            crosser_name:
                The name of the crosser which carried out the crossover.

        """

        self._molecule_record = molecule_record
        self._crosser_name = crosser_name

    def get_molecule_record(self) -> T:
        """
        Get the molecule record produced by the crossover.

        Returns:
            The molecule record.

        """

        return self._molecule_record

    def get_crosser_name(self) -> str:
        """
        Get the name of the crosser which carried out the crossover.

        Returns:
            The name of the crosser.

        """

        return self._crosser_name



from tests.utilities import is_equivalent


def test_mutate(case_data):
    """
    Test :meth:`mutate`.

    Parameters
    ----------
    case_data : :class:`.CaseData`
        A test case. Holds the mutator to test and the correct mutation
        record.

    Returns
    -------
    None : :class:`NoneType`

    """

    _test_mutate(
        mutator=case_data.mutator,
        record=case_data.record,
        mutation_record=case_data.mutation_record,
    )


def _test_mutate(mutator, record, mutation_record):
    """
    Test :meth:`mutate`.

    Parameters
    ----------
    mutator : :class:`.MoleculeMutator`
        The mutator to test.

    record : :class:`.MoleculeRecord`
        The molecule to mutate.

    mutation_record : :class:`.MutationRecord`
        The correct mutation record.

    Returns
    -------
    None : :class:`NoneType`

    """

    result = mutator.mutate(record)
    assert result.get_mutator_name() == mutation_record.get_mutator_name()
    is_equivalent(
        result.get_molecule_record().get_molecule(),
        mutation_record.get_molecule_record().get_molecule(),
    )



import typing
from typing import Any

from .fitness_normalizer import FitnessNormalizer

T = typing.TypeVar("T")


class NullFitnessNormalizer(FitnessNormalizer[T]):
    """
    Does nothing.

    This normalizer just yields the molecule records passed to it,
    without changing them in any way.
    """

    def normalize(self, fitness_values: dict[T, Any]) -> dict[T, Any]:
        return fitness_values



from .utilities import is_clone


def test_clone(edge):
    """
    Test :meth:`.Edge.clone`.

    Parameters
    ----------
    edge : :class:`.Edge`
        The edge to test.

    Returns
    -------
    None : :class:`NoneType`

    """

    clone = edge.clone()
    is_clone(edge, clone)



"""
Dibormo
=======

"""

from .generic_functional_group import GenericFunctionalGroup


class Dibromo(GenericFunctionalGroup):
    """
    Represents a dibromo functional group.

    The structure of the functional group is given by the pseudo-SMILES
    ``[bromine1][atom1][atom2][bromine2]``.

    """

    def __init__(
        self,
        bromine1,
        atom1,
        bromine2,
        atom2,
        bonders,
        deleters,
        placers=None,
    ):
        """
        Initialize a :class:`.Dibromo` instance.

        Parameters
        ----------
        bromine1 : :class:`.Br`
            The ``[bromine1]`` atom.

        atom1 : :class:`.Atom`
            The ``[atom1]`` atom.

        bromine2 : :class:`.Br`
            The ``[bromine2]`` atom.

        atom2 : :class:`.Atom`
            The ``[atom]`` atom.

        bonders : :class:`tuple` of :class:`.Atom`
            The bonder atoms.

        deleters : :class:`tuple` of :class:`.Atom`
            The deleter atoms.

        placers : :class:`tuple` of :class:`.Atom`, optional
            The placer atoms. If ``None`` the `bonders` will be used.

        """

        self._bromine1 = bromine1
        self._atom1 = atom1
        self._bromine2 = bromine2
        self._atom2 = atom2
        super().__init__(
            atoms=(bromine1, atom1, bromine2, atom2),
            bonders=bonders,
            deleters=deleters,
            placers=bonders if placers is None else placers,
        )

    def get_atom1(self):
        """
        Get the ``[atom1]`` atom.

        Returns
        -------
        :class:`.Atom`
            The ``[atom1]`` atom.

        """

        return self._atom1

    def get_bromine1(self):
        """
        Get the ``[bromine1]`` atom.

        Returns
        -------
        :class:`.Br`
            The ``[bromine1]`` atom.

        """

        return self._bromine1

    def get_atom2(self):
        """
        Get the ``[atom2]`` atom.

        Returns
        -------
        :class:`.Atom`
            The ``[atom2]`` atom.

        """

        return self._atom2

    def get_bromine2(self):
        """
        Get the ``[bromine2]`` atom.

        Returns
        -------
        :class:`.Br`
            The ``[bromine2]`` atom.

        """

        return self._bromine2

    def clone(self):
        clone = super().clone()
        clone._atom1 = self._atom1
        clone._bromine1 = self._bromine1
        clone._atom2 = self._atom2
        clone._bromine2 = self._bromine2
        return clone

    def with_atoms(self, atom_map):
        clone = super().with_atoms(atom_map)
        clone._atom1 = atom_map.get(
            self._atom1.get_id(),
            self._atom1,
        )
        clone._bromine1 = atom_map.get(
            self._bromine1.get_id(),
            self._bromine1,
        )
        clone._atom2 = atom_map.get(
            self._atom2.get_id(),
            self._atom2,
        )
        clone._bromine2 = atom_map.get(
            self._bromine2.get_id(),
            self._bromine2,
        )
        return clone

    def __repr__(self):
        return (
            f"{self.__class__.__name__}("
            f"{self._bromine1}, {self._atom1}, {self._bromine2}, "
            f"{self._atom2}, "
            f"bonders={self._bonders}, deleters={self._deleters})"
        )



class CaseData:
    """
    A test case.

    Attributes
    ----------
    crosser : :class:`.MoleculeCrosser`
        The crosser to test.

    records : :class:`tuple` of :class:`.MoleculeRecord`
        The molecules to cross.

    crossover_records : :class:`tuple` of :class:`.CrossoverRecord`
        The correct offspring.

    """

    def __init__(self, crosser, records, crossover_records):
        """
        Initialize a :class:`.CaseData` instance.

        Parameters
        ----------
        crosser : :class:`.MoleculeCrosser`
            The crosser to test.

        records : :class:`tuple` of :class:`.MoleculeRecord`
            The molecules to cross.

        crossover_records : :class:`tuple` :class:`.CrossoverRecord`
            The correct offspring.

        """

        self.crosser = crosser
        self.records = records
        self.crossover_records = crossover_records



import pytest
from pytest_lazyfixture import lazy_fixture

from .case_data import CaseData

# Fixtures must be visible for lazy_fixture() calls.
from .fixtures import *  # noqa


@pytest.fixture(
    params=(
        lazy_fixture("add"),
        lazy_fixture("divide_by_mean"),
        lazy_fixture("multiply"),
        lazy_fixture("null"),
        lazy_fixture("power"),
        lazy_fixture("replace_fitness"),
        lazy_fixture("sequence"),
        lazy_fixture("shift_up"),
        lazy_fixture("sum"),
    ),
)
def case_data(request: pytest.FixtureRequest) -> CaseData:
    return request.param



class CaseData:
    """
    A test case.

    Attributes
    ----------
    vertex : :class:`.Vertex`
        The vertex to test.

    edges : :class:`tuple` of :class:`.Edge`
        The edges connected to :attr:`.vertex`

    building_block : :class:`.BuildingBlock`
        The building block to be placed by :attr:`.vertex`.

    position : :class:`numpy.ndarray`
        The correct centroid of atoms in :attr:`.position_ids` after
        placement.

    alignment_tests : :class:`dict`
        Maps a :class:`callable` to a :class:`numpy.ndarray`.
        Each :class:`callable` takes a singe parameter,
        `building_block`, and returns a :class:`numpy.ndarray`.

        Each :class:`callable` in `alignment_tests` represents a
        test, and the array which it maps to is the correct result
        for that test. If the array returned by the :class:`callable`
        does not match the array it is mapped to, the test will fail.

        The point of these tests is to make sure that
        :attr:`.building_block` is aligned correctly after placement.
        Therefore, alignment tests should return some vector which
        depends on the specific orientation of the building block.

    functional_group_edges : :class:`dict`
        The correct mapping of functional group id to edge id.

    position_ids : :class:`tuple` of :class:`int`
        The ids of atoms which are used to determine if the building
        block was positioned correctly.

    """

    def __init__(
        self,
        vertex,
        edges,
        building_block,
        position,
        alignment_tests,
        functional_group_edges,
        position_ids=None,
    ):
        """
        Initialize a :class:`.CaseData` instance.

        Parameters
        ----------
        vertex : :class:`.Vertex`
            The vertex to test.

        edges : :class:`tuple` of :class:`.Edge`
            The edges connected to `vertex`

        building_block : :class:`.BuildingBlock`
            The building block to be placed by `vertex`.

        position : :class:`numpy.ndarray`
            The correct centroid of atoms in `position_ids` after
            placement.

        alignment_tests : :class:`dict`
            Maps a :class:`callable` to a :class:`numpy.ndarray`.
            Each :class:`callable` takes a singe parameter,
            `building_block`, and returns a :class:`numpy.ndarray`.

            Each :class:`callable` in `alignment_tests` represents a
            test, and the array which it maps to is the correct result
            for that test. If the array returned by the
            :class:`callable` does not match the array it is mapped to,
            the test will fail.

            The point of these tests is to make sure that
            `.building_block` is aligned correctly after
            placement. Therefore, alignment tests should return some
            vector which depends on the specific orientation of the
            building block.

        functional_group_edges : :class:`dict`
            The correct mapping of functional group id to edge id.

        position_ids : :class:`tuple` of :class:`int`, optional
            The ids of atoms which are used to determine if the
            building block was positioned correctly. If ``None``, the
            placer ids will be used.

        """

        if position_ids is None:
            position_ids = tuple(building_block.get_placer_ids())

        self.vertex = vertex
        self.edges = edges
        self.building_block = building_block
        self.position = position
        self.alignment_tests = alignment_tests
        self.functional_group_edges = functional_group_edges
        self.position_ids = position_ids

    def __str__(self):
        return (
            f"CaseData({self.vertex}, {self.edges}, "
            f"{self.building_block}, {self.position}, "
            f"{self.alignment_tests}, {self.functional_group_edges})"
        )

    def __repr__(self):
        return str(self)



import typing

from stk._internal.atom import Atom


class Bond:
    """
    Represents an atomic bond.

    Examples:
        *Changing the Atoms of a Bond*

        You want to substitute the atoms in the bond for other atoms.
        You can do this by using :meth:`.with_atoms` to create a clone
        of the bond, which holds the replacement atoms

        .. testcode:: changing-the-atoms-of-a-bond

            import stk
            bond = stk.Bond(stk.C(0), stk.C(12), 2)
            # Replace C(0) with H(13) but keep C(12).
            clone = bond.with_atoms({0: stk.H(13)})

        .. testcode:: changing-the-atoms-of-a-bond
            :hide:

            assert clone.get_atom1().get_id() == 13
            assert isinstance(clone.get_atom1(), stk.H)
            assert clone.get_atom2().get_id() == 12
            assert isinstance(clone.get_atom2(), stk.C)
    """

    def __init__(
        self,
        atom1: Atom,
        atom2: Atom,
        order: int,
        periodicity: tuple[int, int, int] = (0, 0, 0),
    ) -> None:
        """
        Parameters:

            atom1:
                The first atom in the bond.

            atom2:
                The second atom in the bond.

            order:
                The bond order.

            periodicity:
                The directions across which the bond is periodic. For
                example, ``(1, 0, -1)`` means that when going from
                `atom1` to `atom2` the bond is
                periodic across the x axis in the positive direction,
                is not periodic across the y axis and is periodic
                across the z axis in the negative direction.
        """
        self._atom1 = atom1
        self._atom2 = atom2
        self._order = order
        self._periodicity = periodicity

    def get_atom1(self) -> Atom:
        """
        Get the first atom of the bond.

        Returns:
            The first atom of the bond.
        """
        return self._atom1

    def get_atom2(self) -> Atom:
        """
        Get the second atom of the bond.

        Returns:
            The second atom of the bond.
        """
        return self._atom2

    def get_order(self) -> int:
        """
        Get the bond order of the bond.

        Returns:
            The bond order.
        """
        return self._order

    def get_periodicity(self) -> tuple[int, int, int]:
        """
        Get the periodicity of the bond.

        Returns:
            The directions across which the bond is periodic. For
            example, ``(1, 0, -1)`` means that when going from
            `atom1` to `atom2` the bond is
            periodic across the x axis in the positive direction, is
            not periodic across the y axis and is periodic across the z
            axis in the negative direction.

        """
        return self._periodicity

    def clone(self) -> typing.Self:
        """
        Return a clone.

        Returns:
            Bond: The clone.
        """
        clone = self.__class__.__new__(self.__class__)
        Bond.__init__(
            self=clone,
            atom1=self._atom1,
            atom2=self._atom2,
            order=self._order,
            periodicity=self._periodicity,
        )
        return clone

    def _with_atoms(self, atom_map: dict[int, Atom]) -> typing.Self:
        """
        Modify the bond.

        """
        self._atom1 = atom_map.get(self._atom1.get_id(), self._atom1)
        self._atom2 = atom_map.get(self._atom2.get_id(), self._atom2)
        return self

    def with_atoms(self, atom_map: dict[int, Atom]) -> typing.Self:
        """
        Return a clone holding different atoms.

        Parameters:
            atom_map:
                Maps the id of an atom in the bond to the new atom the
                clone should hold. If the id of an atom in the bond is
                not found in `atom_map`, the atom will not be replaced
                in the clone.

        Returns:
            Bond: The clone.
        """
        return self.clone()._with_atoms(atom_map)

    def with_ids(self, id_map: dict[int, int]) -> typing.Self:
        """
        Return a clone holding different atom ids.

        Parameters:
            id_map:
                Maps the id of an atom in the bond to the new id the
                clone should hold. If the id of an atom in the bond is
                not found in `id_map`, the atom id will not be replaced
                in the clone.

        Returns:
            Bond: The clone.
        """
        return self.clone()._with_ids(id_map)

    def _with_ids(self, id_map: dict[int, int]) -> typing.Self:
        id1 = self._atom1.get_id()
        if id1 in id_map:
            self._atom1 = self._atom1.with_id(id_map[id1])

        id2 = self._atom2.get_id()
        if id2 in id_map:
            self._atom2 = self._atom2.with_id(id_map[id2])

        return self

    def is_periodic(self) -> bool:
        """
        Return ``True`` if the bond is periodic.

        Returns:
            ``True`` if the bond is periodic.
        """
        return any(direction != 0 for direction in self._periodicity)

    def __repr__(self) -> str:
        periodicity = (
            f", periodicity={self._periodicity}" if self.is_periodic() else ""
        )

        cls_name = self.__class__.__name__
        return (
            f"{cls_name}({self._atom1!r}, {self._atom2!r}, "
            f"{self._order}{periodicity})"
        )

    def __str__(self) -> str:
        return repr(self)



import typing
from collections.abc import Callable
from typing import Any

from .fitness_normalizer import FitnessNormalizer

T = typing.TypeVar("T")


class Sum(FitnessNormalizer[T]):
    """
    Combines fitness value components into a single fitness value.

    Examples:

        *Combining Fitness Value Components Into a Single Value*

        Sometimes, your :class:`.FitnessCalculator` will return a fitness
        value which is a :class:`tuple` of multiple numbers. Each number
        represents a different property of the molecule, which
        contributes to the final fitness value. This fitness normalizer
        combines these fitness value components into a single number, by
        taking their sum.

        .. testcode:: combining-fitness-value-components

            import stk

            building_block = stk.BuildingBlock('BrCCBr', stk.BromoFactory())
            record = stk.MoleculeRecord(
                topology_graph=stk.polymer.Linear(
                    building_blocks=[building_block],
                    repeating_unit='A',
                    num_repeating_units=2,
                ),
            )
            fitness_values = {
                record: (1, -2, 3),
            }
            sum_normalizer = stk.Sum()
            normalized_fitness_values = sum_normalizer.normalize(
                fitness_values=fitness_values,
            )
            assert normalized_fitness_values[record] == 2

        *Selectively Normalizing Fitness Values*

        Sometimes, you only want to normalize some members of a population,
        for example if some do not have an assigned fitness value,
        because the fitness calculation failed for whatever reason.
        You can use the `filter` parameter to exclude records from the
        normalization

        .. testcode:: selectively-normalizing-fitness-values

            import stk

            building_block = stk.BuildingBlock('BrCCBr', stk.BromoFactory())
            record1 = stk.MoleculeRecord(
                topology_graph=stk.polymer.Linear(
                    building_blocks=[building_block],
                    repeating_unit='A',
                    num_repeating_units=2,
                ),
            )
            record2 = stk.MoleculeRecord(
                topology_graph=stk.polymer.Linear(
                    building_blocks=[building_block],
                    repeating_unit='A',
                    num_repeating_units=2,
                ),
            )
            fitness_values = {
                record1: (1, -2, 3),
                record2: None,
            }
            normalizer = stk.Sum(
                # Only normalize values which are not None.
                filter=lambda fitness_values, record:
                    fitness_values[record] is not None,
            )
            normalized_fitness_values = normalizer.normalize(fitness_values)
            assert normalized_fitness_values[record1] == 2
            assert normalized_fitness_values[record2] is None
    """

    def __init__(
        self,
        filter: Callable[
            [dict[T, Any], T], bool
        ] = lambda fitness_values, record: True,
    ) -> None:
        """
        Parameters:
            filter:
                A function wihch returns ``True`` or ``False``. Only
                molecules which return ``True`` will have fitness values
                normalized. By default, all molecules will have fitness
                values normalized.
                The instance passed to the `fitness_values` argument of
                :meth:`.normalize` is passed as the first argument, while
                the second argument will be passed every
                :class:`.MoleculeRecord` in it, one at a time.
        """
        self._filter = filter

    def normalize(self, fitness_values: dict[T, Any]) -> dict[T, Any]:
        return {
            record: sum(fitness_value)
            if self._filter(fitness_values, record)
            else fitness_value
            for record, fitness_value in fitness_values.items()
        }



import pytest
import rdkit.Chem.AllChem as rdkit
import stk

from ..case_data import CaseData


def _get_rdkit_molecule() -> rdkit.Mol:
    molecule = rdkit.MolFromSmiles("Br[C+2][C+2]Br")
    rdkit.EmbedMolecule(molecule, rdkit.ETKDGv2())
    return molecule


@pytest.fixture(
    scope="session",
    params=(
        lambda: CaseData(
            building_block=stk.BuildingBlock.init_from_rdkit_mol(
                molecule=_get_rdkit_molecule(),
            ),
            functional_groups=(),
            core_atom_ids=(0, 1, 2, 3),
            placer_ids=(0, 1, 2, 3),
        ),
        lambda: CaseData(
            building_block=stk.BuildingBlock.init_from_rdkit_mol(
                molecule=_get_rdkit_molecule(),
                functional_groups=[stk.BromoFactory()],
            ),
            functional_groups=(
                stk.Bromo(
                    bromine=stk.Br(0),
                    atom=stk.C(1, 2),
                    bonders=(stk.C(1, 2),),
                    deleters=(stk.Br(0),),
                ),
                stk.Bromo(
                    bromine=stk.Br(3),
                    atom=stk.C(2, 2),
                    bonders=(stk.C(2, 2),),
                    deleters=(stk.Br(3),),
                ),
            ),
            core_atom_ids=(1, 2),
            placer_ids=(1, 2),
        ),
        lambda: CaseData(
            building_block=stk.BuildingBlock.init_from_rdkit_mol(
                molecule=_get_rdkit_molecule(),
                placer_ids=(1, 2),
            ),
            functional_groups=(),
            core_atom_ids=(0, 1, 2, 3),
            placer_ids=(1, 2),
        ),
        lambda: CaseData(
            building_block=stk.BuildingBlock.init_from_rdkit_mol(
                molecule=_get_rdkit_molecule(),
                functional_groups=[stk.BromoFactory()],
                placer_ids=(0, 3),
            ),
            functional_groups=(
                stk.Bromo(
                    bromine=stk.Br(0),
                    atom=stk.C(1, 2),
                    bonders=(stk.C(1, 2),),
                    deleters=(stk.Br(0),),
                ),
                stk.Bromo(
                    bromine=stk.Br(3),
                    atom=stk.C(2, 2),
                    bonders=(stk.C(2, 2),),
                    deleters=(stk.Br(3),),
                ),
            ),
            core_atom_ids=(1, 2),
            placer_ids=(0, 3),
        ),
        lambda: CaseData(
            building_block=stk.BuildingBlock.init_from_rdkit_mol(
                molecule=_get_rdkit_molecule(),
                functional_groups=[stk.IodoFactory()],
            ),
            functional_groups=(),
            core_atom_ids=(0, 1, 2, 3),
            placer_ids=(0, 1, 2, 3),
        ),
    ),
)
def init_from_rdkit_mol(request) -> CaseData:
    return request.param()



from __future__ import annotations

import typing
from collections import Counter, abc, defaultdict
from functools import partial

from stk._internal.building_block import BuildingBlock
from stk._internal.optimizers.null import NullOptimizer
from stk._internal.optimizers.optimizer import Optimizer
from stk._internal.reaction_factories.generic_reaction_factory import (
    GenericReactionFactory,
)
from stk._internal.reaction_factories.reaction_factory import ReactionFactory
from stk._internal.topology_graphs.edge import Edge
from stk._internal.topology_graphs.topology_graph.topology_graph import (
    TopologyGraph,
)
from stk._internal.topology_graphs.vertex import Vertex

from .cage_construction_state import _CageConstructionState
from .vertices import UnaligningVertex, _CageVertex


class UnoccupiedVertexError(Exception):
    """
    When a cage vertex is not occupied by a building block.

    """

    pass


class OverlyOccupiedVertexError(Exception):
    """
    When a cage vertex is occupied by more than one building block.

    """

    pass


class Cage(TopologyGraph):
    """
    Represents a cage topology graph.

    Notes:

        Cage topologies are added by creating a subclass, which defines
        the :attr:`_vertex_prototypes` and :attr:`_edge_prototypes`
        class attributes.

    .. _cage-topology-graph-examples:

    Examples:

        *Subclass Implementation*

        The source code of the subclasses, listed in
        :mod:`~.cage.cage`, can serve as good examples.

        *Basic Construction*

        :class:`.Cage` instances can be made by providing the building
        block molecules only (using :class:`.FourPlusSix` as an
        example)

        .. testcode:: basic-construction

            import stk

            bb1 = stk.BuildingBlock(
                smiles='NCCN',
                functional_groups=[stk.PrimaryAminoFactory()],
            )
            bb2 = stk.BuildingBlock(
                smiles='O=CC(C=O)C=O',
                functional_groups=[stk.AldehydeFactory()],
            )
            cage = stk.ConstructedMolecule(
                topology_graph=stk.cage.FourPlusSix((bb1, bb2)),
            )

        .. moldoc::

            import moldoc.molecule as molecule
            import stk

            bb1 = stk.BuildingBlock(
                smiles='NCCN',
                functional_groups=[stk.PrimaryAminoFactory()],
            )
            bb2 = stk.BuildingBlock(
                smiles='O=CC(C=O)C=O',
                functional_groups=[stk.AldehydeFactory()],
            )
            cage = stk.ConstructedMolecule(
                topology_graph=stk.cage.FourPlusSix((bb1, bb2)),
            )

            moldoc_display_molecule = molecule.Molecule(
                atoms=(
                    molecule.Atom(
                        atomic_number=atom.get_atomic_number(),
                        position=position,
                    ) for atom, position in zip(
                        cage.get_atoms(),
                        cage.get_position_matrix(),
                    )
                ),
                bonds=(
                    molecule.Bond(
                        atom1_id=bond.get_atom1().get_id(),
                        atom2_id=bond.get_atom2().get_id(),
                        order=bond.get_order(),
                    ) for bond in cage.get_bonds()
                ),
            )

        *Suggested Optimization*

        For :class:`.Cage` topologies, it is recommend to use the
        :class:`.MCHammer` optimizer.
        However, for cages formed from highly unsymmetrical building
        blocks, it is recommend to use the simplified
        :class:`.Collapser` optimizer.


        .. testcode:: suggested-optimization

            import stk

            bb1 = stk.BuildingBlock(
                smiles='NCCN',
                functional_groups=[stk.PrimaryAminoFactory()],
            )
            bb2 = stk.BuildingBlock(
                smiles='O=CC(C=O)C=O',
                functional_groups=[stk.AldehydeFactory()],
            )

            cage = stk.ConstructedMolecule(
                topology_graph=stk.cage.FourPlusSix(
                    building_blocks=(bb1, bb2),
                    optimizer=stk.MCHammer(),
                ),
            )

        .. moldoc::

            import moldoc.molecule as molecule
            import stk

            bb1 = stk.BuildingBlock(
                smiles='NCCN',
                functional_groups=[stk.PrimaryAminoFactory()],
            )
            bb2 = stk.BuildingBlock(
                smiles='O=CC(C=O)C=O',
                functional_groups=[stk.AldehydeFactory()],
            )

            cage = stk.ConstructedMolecule(
                topology_graph=stk.cage.FourPlusSix(
                    building_blocks=(bb1, bb2),
                    optimizer=stk.MCHammer(),
                ),
            )

            moldoc_display_molecule = molecule.Molecule(
                atoms=(
                    molecule.Atom(
                        atomic_number=atom.get_atomic_number(),
                        position=position,
                    ) for atom, position in zip(
                        cage.get_atoms(),
                        cage.get_position_matrix(),
                    )
                ),
                bonds=(
                    molecule.Bond(
                        atom1_id=bond.get_atom1().get_id(),
                        atom2_id=bond.get_atom2().get_id(),
                        order=bond.get_order(),
                    ) for bond in cage.get_bonds()
                ),
            )

        *Structural Isomer Construction*

        Different structural isomers of cages can be made by using the
        `vertex_alignments` optional parameter

        .. testcode:: structural-isomer-construction

            import stk

            bb1 = stk.BuildingBlock(
                smiles='NCCN',
                functional_groups=[stk.PrimaryAminoFactory()],
            )
            bb2 = stk.BuildingBlock(
                smiles='O=CC(C=O)C=O',
                functional_groups=[stk.AldehydeFactory()],
            )

            cage = stk.ConstructedMolecule(
                topology_graph=stk.cage.FourPlusSix(
                    building_blocks=(bb1, bb2),
                    vertex_alignments={0: 1, 1: 1, 2: 2},
                ),
            )

        The parameter maps the id of a vertex to a number
        between 0 (inclusive) and the number of edges the vertex is
        connected to (exclusive). So a vertex connected to three edges
        can be mapped to ``0``, ``1`` or ``2``.

        By changing which edge each vertex is aligned with, a different
        structural isomer of the cage can be formed.

        *Multi-Building Block Cage Construction*

        You can also build cages with multiple building blocks, but,
        if you have multiple building blocks with the same number
        of functional groups, you have to assign each building block to
        the vertex you want to place it on

        .. testcode:: multi-building-block-cage-construction

            import stk

            bb1 = stk.BuildingBlock(
                smiles='O=CC(C=O)C=O',
                functional_groups=[stk.AldehydeFactory()],
            )
            bb2 = stk.BuildingBlock(
                smiles='O=CC(Cl)(C=O)C=O',
                functional_groups=[stk.AldehydeFactory()],
            )
            bb3 = stk.BuildingBlock(
                smiles='NCCN',
                functional_groups=[stk.PrimaryAminoFactory()],
            )
            bb4 = stk.BuildingBlock(
                smiles='NCC(Cl)N',
                functional_groups=[stk.PrimaryAminoFactory()],
            )
            bb5 = stk.BuildingBlock(
                smiles='NCCCCN',
                functional_groups=[stk.PrimaryAminoFactory()],
            )

            cage1 = stk.ConstructedMolecule(
                topology_graph=stk.cage.FourPlusSix(
                    building_blocks={
                        bb1: range(2),
                        bb2: (2, 3),
                        bb3: 4,
                        bb4: 5,
                        bb5: range(6, 10),
                    },
                ),
            )

        .. moldoc::

            import moldoc.molecule as molecule
            import stk

            bb1 = stk.BuildingBlock(
                smiles='O=CC(C=O)C=O',
                functional_groups=[stk.AldehydeFactory()],
            )
            bb2 = stk.BuildingBlock(
                smiles='O=CC(Cl)(C=O)C=O',
                functional_groups=[stk.AldehydeFactory()],
            )
            bb3 = stk.BuildingBlock(
                smiles='NCCN',
                functional_groups=[stk.PrimaryAminoFactory()],
            )
            bb4 = stk.BuildingBlock(
                smiles='NCC(Cl)N',
                functional_groups=[stk.PrimaryAminoFactory()],
            )
            bb5 = stk.BuildingBlock(
                smiles='NCCCCN',
                functional_groups=[stk.PrimaryAminoFactory()],
            )

            cage = stk.ConstructedMolecule(
                topology_graph=stk.cage.FourPlusSix(
                    building_blocks={
                        bb1: range(2),
                        bb2: (2, 3),
                        bb3: 4,
                        bb4: 5,
                        bb5: range(6, 10),
                    },
                ),
            )

            moldoc_display_molecule = molecule.Molecule(
                atoms=(
                    molecule.Atom(
                        atomic_number=atom.get_atomic_number(),
                        position=position,
                    ) for atom, position in zip(
                        cage.get_atoms(),
                        cage.get_position_matrix(),
                    )
                ),
                bonds=(
                    molecule.Bond(
                        atom1_id=bond.get_atom1().get_id(),
                        atom2_id=bond.get_atom2().get_id(),
                        order=bond.get_order(),
                    ) for bond in cage.get_bonds()
                ),
            )


        You can combine this with the `vertex_alignments` parameter

        .. testcode:: multi-building-block-cage-construction

            cage2 = stk.ConstructedMolecule(
                topology_graph=stk.cage.FourPlusSix(
                    building_blocks={
                        bb1: range(2),
                        bb2: (2, 3),
                        bb3: 4,
                        bb4: 5,
                        bb5: range(6, 10),
                    },
                    vertex_alignments={5: 1},
                ),
            )

        .. moldoc::

            import moldoc.molecule as molecule
            import stk

            bb1 = stk.BuildingBlock(
                smiles='O=CC(C=O)C=O',
                functional_groups=[stk.AldehydeFactory()],
            )
            bb2 = stk.BuildingBlock(
                smiles='O=CC(Cl)(C=O)C=O',
                functional_groups=[stk.AldehydeFactory()],
            )
            bb3 = stk.BuildingBlock(
                smiles='NCCN',
                functional_groups=[stk.PrimaryAminoFactory()],
            )
            bb4 = stk.BuildingBlock(
                smiles='NCC(Cl)N',
                functional_groups=[stk.PrimaryAminoFactory()],
            )
            bb5 = stk.BuildingBlock(
                smiles='NCCCCN',
                functional_groups=[stk.PrimaryAminoFactory()],
            )

            cage = stk.ConstructedMolecule(
                topology_graph=stk.cage.FourPlusSix(
                    building_blocks={
                        bb1: range(2),
                        bb2: (2, 3),
                        bb3: 4,
                        bb4: 5,
                        bb5: range(6, 10),
                    },
                    vertex_alignments={5: 1},
                ),
            )

            moldoc_display_molecule = molecule.Molecule(
                atoms=(
                    molecule.Atom(
                        atomic_number=atom.get_atomic_number(),
                        position=position,
                    ) for atom, position in zip(
                        cage.get_atoms(),
                        cage.get_position_matrix(),
                    )
                ),
                bonds=(
                    molecule.Bond(
                        atom1_id=bond.get_atom1().get_id(),
                        atom2_id=bond.get_atom2().get_id(),
                        order=bond.get_order(),
                    ) for bond in cage.get_bonds()
                ),
            )

        *Metal-Organic Cage Construction*

        A series of common metal-organic cage topologies are provided
        and can be constructed in the same way as other :class:`.Cage`
        instances using metal atoms and :class:`DativeReactionFactory`
        instances to produce metal-ligand bonds. Each metal topology
        has specific vertices reserved for the metal atoms or
        complexes, which are listed in their documentation.

        .. testcode:: metal-organic-cage-construction

            import stk

            # Produce a Pd+2 atom with 4 functional groups.
            palladium_atom = stk.BuildingBlock(
                smiles='[Pd+2]',
                functional_groups=(
                    stk.SingleAtom(stk.Pd(0, charge=2))
                    for i in range(4)
                ),
                position_matrix=[[0., 0., 0.]],
            )

            # Build a building block with two functional groups using
            # the SmartsFunctionalGroupFactory.
            bb1 = stk.BuildingBlock(
                smiles=(
                    'C1=NC=CC(C2=CC=CC(C3=C'
                    'C=NC=C3)=C2)=C1'
                ),
                functional_groups=[
                    stk.SmartsFunctionalGroupFactory(
                        smarts='[#6]~[#7X2]~[#6]',
                        bonders=(1, ),
                        deleters=(),
                    ),
                ],
            )

            cage1 = stk.ConstructedMolecule(
                stk.cage.M2L4Lantern(
                    building_blocks=(palladium_atom, bb1),
                    # Ensure that bonds between the
                    # GenericFunctionalGroups of the ligand and the
                    # SingleAtom functional groups of the metal are
                    # dative.
                    reaction_factory=stk.DativeReactionFactory(
                        stk.GenericReactionFactory(
                            bond_orders={
                                frozenset({
                                    stk.GenericFunctionalGroup,
                                    stk.SingleAtom,
                                }): 9,
                            },
                        ),
                    ),
                ),
            )

        .. moldoc::

            import moldoc.molecule as molecule
            import stk

            palladium_atom = stk.BuildingBlock(
                smiles='[Pd+2]',
                functional_groups=(
                    stk.SingleAtom(stk.Pd(0, charge=2))
                    for i in range(4)
                ),
                position_matrix=[[0., 0., 0.]],
            )

            bb1 = stk.BuildingBlock(
                smiles=(
                    'C1=NC=CC(C2=CC=CC(C3=C'
                    'C=NC=C3)=C2)=C1'
                ),
                functional_groups=[
                    stk.SmartsFunctionalGroupFactory(
                        smarts='[#6]~[#7X2]~[#6]',
                        bonders=(1, ),
                        deleters=(),
                    ),
                ],
            )

            cage = stk.ConstructedMolecule(
                stk.cage.M2L4Lantern(
                    building_blocks=(palladium_atom, bb1),
                    reaction_factory=stk.DativeReactionFactory(
                        stk.GenericReactionFactory(
                            bond_orders={
                                # Use bond order of 1 here so that the
                                # rendering does not show a bond order
                                # of 9.
                                frozenset({
                                    stk.GenericFunctionalGroup,
                                    stk.SingleAtom,
                                }): 1,
                            },
                        ),
                    ),
                ),
            )

            moldoc_display_molecule = molecule.Molecule(
                atoms=(
                    molecule.Atom(
                        atomic_number=atom.get_atomic_number(),
                        position=position,
                    ) for atom, position in zip(
                        cage.get_atoms(),
                        cage.get_position_matrix(),
                    )
                ),
                bonds=(
                    molecule.Bond(
                        atom1_id=bond.get_atom1().get_id(),
                        atom2_id=bond.get_atom2().get_id(),
                        order=bond.get_order(),
                    ) for bond in cage.get_bonds()
                ),
            )

        *Controlling Metal-Complex Stereochemistry*

        When building metal-organic cages from octahedral metals, i.e.
        Fe(II), the stereochemistry of the metal centre can be
        important. Maintaining that stereochemistry around specific
        metal centres during :class:`.Cage` construction is difficult,
        so an alternative route to these types of structures can be
        taken. Firstly, you would construct a :class:`.MetalComplex`
        instance with the appropriate stereochemistry and dummy
        reactive groups (bromine in the following example)

        .. testcode:: controlling-metal-complex-stereochemistry

            import stk

            # Produce a Fe+2 atom with 6 functional groups.
            iron_atom = stk.BuildingBlock(
                smiles='[Fe+2]',
                functional_groups=(
                    stk.SingleAtom(stk.Fe(0, charge=2))
                    for i in range(6)
                ),
                position_matrix=[[0, 0, 0]],
            )

            # Define coordinating ligand with dummy bromine groups and
            # metal coordinating functional groups.
            bb2 = stk.BuildingBlock(
                smiles='C1=NC(C=NBr)=CC=C1',
                functional_groups=[
                    stk.SmartsFunctionalGroupFactory(
                        smarts='[#6]~[#7X2]~[#35]',
                        bonders=(1, ),
                        deleters=(),
                    ),
                    stk.SmartsFunctionalGroupFactory(
                        smarts='[#6]~[#7X2]~[#6]',
                        bonders=(1, ),
                        deleters=(),
                    ),
                ],
            )

            # Build iron complex with delta stereochemistry.
            iron_oct_delta = stk.ConstructedMolecule(
                topology_graph=stk.metal_complex.OctahedralDelta(
                    metals=iron_atom,
                    ligands=bb2,
                ),
            )

        .. moldoc::

            import moldoc.molecule as molecule
            import stk

            iron_atom = stk.BuildingBlock(
                smiles='[Fe+2]',
                functional_groups=(
                    stk.SingleAtom(stk.Fe(0, charge=2))
                    for i in range(6)
                ),
                position_matrix=[[0, 0, 0]],
            )

            bb2 = stk.BuildingBlock(
                smiles='C1=NC(C=NBr)=CC=C1',
                functional_groups=[
                    stk.SmartsFunctionalGroupFactory(
                        smarts='[#6]~[#7X2]~[#35]',
                        bonders=(1, ),
                        deleters=(),
                    ),
                    stk.SmartsFunctionalGroupFactory(
                        smarts='[#6]~[#7X2]~[#6]',
                        bonders=(1, ),
                        deleters=(),
                    ),
                ],
            )

            complex = stk.ConstructedMolecule(
                topology_graph=stk.metal_complex.OctahedralDelta(
                    metals=iron_atom,
                    ligands=bb2,
                ),
            )

            moldoc_display_molecule = molecule.Molecule(
                atoms=(
                    molecule.Atom(
                        atomic_number=atom.get_atomic_number(),
                        position=position,
                    ) for atom, position in zip(
                        complex.get_atoms(),
                        complex.get_position_matrix(),
                    )
                ),
                bonds=(
                    molecule.Bond(
                        atom1_id=bond.get_atom1().get_id(),
                        atom2_id=bond.get_atom2().get_id(),
                        order=(
                            1
                            if bond.get_order() == 9
                            else bond.get_order()
                        ),
                    ) for bond in complex.get_bonds()
                ),
            )

        Then the metal complexes can be placed on the appropriate
        :class:`.Cage` topology to produce a structure with the desired
        stereochemistry at all metal centres.

        .. testcode:: controlling-metal-complex-stereochemistry

            # Assign Bromo functional groups to the metal complex.
            iron_oct_delta = stk.BuildingBlock.init_from_molecule(
                molecule=iron_oct_delta,
                functional_groups=[stk.BromoFactory()],
            )

            # Define spacer building block.
            bb3 = stk.BuildingBlock(
                smiles=(
                    'C1=CC(C2=CC=C(Br)C=C2)=C'
                    'C=C1Br'
                ),
                functional_groups=[stk.BromoFactory()],
            )

            # Build an M4L6 Tetrahedron with a spacer.
            cage2 = stk.ConstructedMolecule(
                topology_graph=stk.cage.M4L6TetrahedronSpacer(
                    building_blocks=(
                        iron_oct_delta,
                        bb3,
                    ),
                ),
            )

        .. moldoc::

            import moldoc.molecule as molecule
            import stk

            iron_atom = stk.BuildingBlock(
                smiles='[Fe+2]',
                functional_groups=(
                    stk.SingleAtom(stk.Fe(0, charge=2))
                    for i in range(6)
                ),
                position_matrix=[[0, 0, 0]],
            )

            bb2 = stk.BuildingBlock(
                smiles='C1=NC(C=NBr)=CC=C1',
                functional_groups=[
                    stk.SmartsFunctionalGroupFactory(
                        smarts='[#6]~[#7X2]~[#35]',
                        bonders=(1, ),
                        deleters=(),
                    ),
                    stk.SmartsFunctionalGroupFactory(
                        smarts='[#6]~[#7X2]~[#6]',
                        bonders=(1, ),
                        deleters=(),
                    ),
                ],
            )

            iron_oct_delta = stk.ConstructedMolecule(
                topology_graph=stk.metal_complex.OctahedralDelta(
                    metals=iron_atom,
                    ligands=bb2,
                ),
            )

            iron_oct_delta = stk.BuildingBlock.init_from_molecule(
                molecule=iron_oct_delta,
                functional_groups=[stk.BromoFactory()],
            )

            bb3 = stk.BuildingBlock(
                smiles=(
                    'C1=CC(C2=CC=C(Br)C=C2)=C'
                    'C=C1Br'
                ),
                functional_groups=[stk.BromoFactory()],
            )

            cage = stk.ConstructedMolecule(
                topology_graph=stk.cage.M4L6TetrahedronSpacer(
                    building_blocks=(
                        iron_oct_delta,
                        bb3,
                    ),
                ),
            )

            moldoc_display_molecule = molecule.Molecule(
                atoms=(
                    molecule.Atom(
                        atomic_number=atom.get_atomic_number(),
                        position=position,
                    ) for atom, position in zip(
                        cage.get_atoms(),
                        cage.get_position_matrix(),
                    )
                ),
                bonds=(
                    molecule.Bond(
                        atom1_id=bond.get_atom1().get_id(),
                        atom2_id=bond.get_atom2().get_id(),
                        order=(
                            1
                            if bond.get_order() == 9
                            else bond.get_order()
                        ),
                    ) for bond in cage.get_bonds()
                ),
            )

        *Aligning Metal Complex Building Blocks*

        When building metal-organic cages from metal complex
        building blocks, it is common that
        the metal complex :class:`.BuildingBlock` will have
        multiple functional groups, but that those functional groups
        are overlapping. This means that some of its atoms appear in
        multiple functional groups. A difficulty arises when the
        atom shared between the functional groups is a *placer* atom.

        *Placer* atoms are used to align building blocks, so that
        they have an appropriate orientation in the final topology.
        If there is only one *placer* atom, no alignment can be made,
        as no vector running between *placer* atoms can be defined,
        and used for the alignment of the :class:`.BuildingBlock`.

        By default, :mod:`stk` may create overlapping functional
        groups, which may lead to a lack of an appropriate number
        of *placer* atoms, leading to a :class:`.BuildingBlock`
        being unaligned. However, the user can manually set the
        *placer* atoms of functional groups, so that not all of the
        *placer* atoms appear in multiple functional groups, which
        leads to proper alignment.

        First we build a metal complex

        .. testcode:: aligning-metal-complex-building-blocks

            import stk

            metal_atom = stk.BuildingBlock(
                smiles='[Pd+2]',
                functional_groups=(
                    stk.SingleAtom(stk.Pd(0, charge=2))
                    for i in range(4)
                ),
                position_matrix=[[0., 0., 0.]],
            )

            ligand = stk.BuildingBlock(
                smiles='NCCN',
                functional_groups=[
                    stk.SmartsFunctionalGroupFactory(
                        smarts='[#7]~[#6]',
                        bonders=(0, ),
                        deleters=(),
                    ),
                ],
            )

            metal_complex = stk.ConstructedMolecule(
                topology_graph=stk.metal_complex.CisProtectedSquarePlanar(
                    metals=metal_atom,
                    ligands=ligand,
                ),
            )

        Next, we convert the metal complex into a
        :class:`.BuildingBlock`, taking care to define functional
        groups which do not have overlapping *placer* atoms

        .. testcode:: aligning-metal-complex-building-blocks

            metal_complex = stk.BuildingBlock.init_from_molecule(
                molecule=metal_complex,
                functional_groups=[
                    stk.SmartsFunctionalGroupFactory(
                        smarts='[Pd]~[#7]',
                        bonders=(0, ),
                        deleters=(),
                        # The nitrogen atom will be different
                        # for each functional group.
                        placers=(0, 1),
                    ),
                ],
            )

        We load in the organic linker of the cage as normal

        .. testcode:: aligning-metal-complex-building-blocks

            linker = stk.BuildingBlock(
                smiles='C1=NC=CC(C2=CC=NC=C2)=C1',
                functional_groups=[
                    stk.SmartsFunctionalGroupFactory(
                        smarts='[#6]~[#7X2]~[#6]',
                        bonders=(1, ),
                        deleters=(),
                    ),
                ],
            )

        And finally, we build the cage with a
        :class:`DativeReactionFactory` instance to produce dative
        bonds.

        .. testcode:: aligning-metal-complex-building-blocks

            cage = stk.ConstructedMolecule(
                topology_graph=stk.cage.M4L4Square(
                    corners=metal_complex,
                    linkers=linker,
                    reaction_factory=stk.DativeReactionFactory(
                        stk.GenericReactionFactory(
                            bond_orders={
                                frozenset({
                                    stk.GenericFunctionalGroup,
                                    stk.GenericFunctionalGroup,
                                }): 9,
                            },
                        ),
                    ),
                ),
            )

    """

    _vertex_degrees: typing.ClassVar[dict[int, int]]
    _vertex_prototypes: typing.ClassVar[tuple[_CageVertex, ...]]
    _edge_prototypes: typing.ClassVar[tuple[Edge, ...]]
    _vertices_of_degree: typing.ClassVar[dict[int, set[int]]]

    def __init_subclass__(cls, **kwargs) -> None:
        cls._vertex_degrees = Counter(
            vertex_id
            for edge in cls._edge_prototypes
            for vertex_id in edge.get_vertex_ids()
        )
        cls._vertices_of_degree = defaultdict(set)
        for vertex_id, degree in cls._vertex_degrees.items():
            cls._vertices_of_degree[degree].add(vertex_id)

    def __init__(
        self,
        building_blocks: typing.Union[
            typing.Iterable[BuildingBlock],
            dict[BuildingBlock, tuple[int, ...]],
        ],
        vertex_alignments: typing.Optional[dict[int, int]] = None,
        reaction_factory: ReactionFactory = GenericReactionFactory(),
        num_processes: int = 1,
        optimizer: Optimizer = NullOptimizer(),
    ) -> None:
        """
        Initialize a :class:`.Cage`.

        Parameters:

            building_blocks:
                Can be a :class:`iterable` of :class:`.BuildingBlock`
                instances, which should be placed on the topology
                graph.

                Can also be a :class:`dict` which maps the
                :class:`.BuildingBlock` instances to the ids of the
                vertices it should be placed on. A :class:`dict` is
                required when there are multiple building blocks with
                the same number of functional groups, because in this
                case the desired placement is ambiguous.

            vertex_alignments:
                A mapping from the id of a :class:`.Vertex`
                to an :class:`.Edge` connected to it.
                The :class:`.Edge` is used to align the first
                :class:`.FunctionalGroup` of a :class:`.BuildingBlock`
                placed on that vertex. Only vertices which need to have
                their default edge changed need to be present in the
                :class:`dict`. If ``None`` then the default edge is
                used for each vertex. Changing which :class:`.Edge` is
                used will mean that the topology graph represents
                different structural isomers. The edge is referred to
                by a number between ``0`` (inclusive) and the number of
                edges the vertex is connected to (exclusive).

            reaction_factory:
                The reaction factory to use for creating bonds between
                building blocks.

            num_processes:
                The number of parallel processes to create during
                :meth:`construct`.

            optimizer:
                Used to optimize the structure of the constructed
                molecule.

        Raises:

            :class:`AssertionError`
                If the any building block does not have a
                valid number of functional groups.

            :class:`ValueError`
                If the there are multiple building blocks with the
                same number of functional_groups in `building_blocks`,
                and they are not explicitly assigned to vertices. The
                desired placement of building blocks is ambiguous in
                this case.

            :class:`~.cage.UnoccupiedVertexError`
                If a vertex of the cage topology graph does not have a
                building block placed on it.

            :class:`~.cage.OverlyOccupiedVertexError`
                If a vertex of the cage topology graph has more than
                one building block placed on it.

        """

        building_block_vertices = self._normalize_building_blocks(
            building_blocks=building_blocks,
        )
        self._vertex_alignments = (
            dict(vertex_alignments) if vertex_alignments is not None else {}
        )
        building_block_vertices = self._with_unaligning_vertices(
            building_block_vertices=building_block_vertices,
        )
        building_block_vertices = self._assign_aligners(
            building_block_vertices=building_block_vertices,
            vertex_alignments=self._vertex_alignments,
        )
        self._check_building_block_vertices(building_block_vertices)
        super().__init__(
            building_block_vertices=typing.cast(
                dict[BuildingBlock, abc.Sequence[Vertex]],
                building_block_vertices,
            ),
            edges=self._edge_prototypes,
            reaction_factory=reaction_factory,
            construction_stages=tuple(
                partial(self._has_degree, degree)
                for degree in sorted(self._vertices_of_degree, reverse=True)
            ),
            num_processes=num_processes,
            optimizer=optimizer,
            edge_groups=None,
        )

    @classmethod
    def _normalize_building_blocks(
        cls,
        building_blocks: typing.Union[
            typing.Iterable[BuildingBlock],
            dict[BuildingBlock, tuple[int, ...]],
        ],
    ) -> dict[BuildingBlock, abc.Sequence[_CageVertex]]:
        # Use tuple here because it prints nicely.
        allowed_degrees = tuple(cls._vertices_of_degree.keys())
        if isinstance(building_blocks, dict):
            for building_block in building_blocks:
                assert (
                    building_block.get_num_functional_groups()
                    in cls._vertices_of_degree.keys()
                ), (
                    "The number of functional groups in "
                    f"{building_block} needs to be one of "
                    f"{allowed_degrees}, but is "
                    "currently "
                    f"{building_block.get_num_functional_groups()}."
                )
            return {
                building_block: tuple(cls._get_vertices(ids))
                for building_block, ids in building_blocks.items()
            }

        else:
            return cls._get_building_block_vertices(
                building_blocks=building_blocks,
            )

    @staticmethod
    def _with_unaligning_vertices(
        building_block_vertices: dict[
            BuildingBlock, abc.Sequence[_CageVertex]
        ],
    ) -> dict[BuildingBlock, abc.Sequence[_CageVertex]]:
        clone = dict(building_block_vertices)
        for building_block, vertices in clone.items():
            # Building blocks with 1 placer, cannot be aligned and
            # must therefore use an UnaligningVertex.
            if building_block.get_num_placers() == 1:
                clone[building_block] = tuple(
                    UnaligningVertex(
                        id=v.get_id(),
                        position=v.get_position(),
                        use_neighbor_placement=(v.use_neighbor_placement()),
                        aligner_edge=v.get_aligner_edge(),
                    )
                    for v in vertices
                )

        return clone

    @classmethod
    def _assign_aligners(
        cls,
        building_block_vertices: dict[
            BuildingBlock, abc.Sequence[_CageVertex]
        ],
        vertex_alignments: dict[int, int],
    ) -> dict[BuildingBlock, abc.Sequence[_CageVertex]]:
        def with_aligner(vertex: _CageVertex) -> _CageVertex:
            return vertex.with_aligner_edge(
                aligner_edge=vertex_alignments.get(vertex.get_id(), 0),
            )

        return {
            building_block: tuple(map(with_aligner, vertices))
            for building_block, vertices in building_block_vertices.items()
        }

    @classmethod
    def _check_building_block_vertices(
        cls,
        building_block_vertices: dict[
            BuildingBlock, abc.Sequence[_CageVertex]
        ],
    ) -> None:
        unassigned_ids = set(
            vertex.get_id() for vertex in cls._vertex_prototypes
        )
        assigned_ids = set()
        vertices = (
            vertex
            for vertices_ in building_block_vertices.values()
            for vertex in vertices_
        )
        for vertex in vertices:
            if vertex.get_id() in assigned_ids:
                raise OverlyOccupiedVertexError(
                    f"Vertex {vertex.get_id()} has multiple building "
                    "blocks placed on it."
                )
            assigned_ids.add(vertex.get_id())
            unassigned_ids.remove(vertex.get_id())

        if unassigned_ids:
            raise UnoccupiedVertexError(
                "The following vertices are unoccupied " f"{unassigned_ids}."
            )

    def clone(self) -> Cage:
        clone = self._clone()
        clone._vertex_alignments = dict(self._vertex_alignments)
        return clone

    @classmethod
    def _get_vertices(
        cls,
        vertex_ids: typing.Union[int, typing.Iterable[int]],
    ) -> typing.Iterator[_CageVertex]:
        """
        Yield vertex prototypes.

        Parameters:

            vertex_ids:
                The ids of the vertices to yield.

        Yields:

            A vertex prototype of the topology graph.

        """

        if isinstance(vertex_ids, int):
            vertex_ids = (vertex_ids,)

        for vertex_id in vertex_ids:
            yield cls._vertex_prototypes[vertex_id]

    def _has_degree(
        self,
        degree: int,
        vertex: Vertex,
    ) -> bool:
        """
        Check if `vertex` has a degree of `degree`.

        Parameters:

            degree:
                The degree in question.

            vertex:
                The vertex in question.

        Returns:

            ``True`` if `vertex` has a degree of `degree`.

        """

        return vertex.get_id() in self._vertices_of_degree[degree]

    @classmethod
    def _get_building_block_vertices(
        cls,
        building_blocks: typing.Iterable[BuildingBlock],
    ) -> dict[BuildingBlock, abc.Sequence[_CageVertex]]:
        """
        Map building blocks to the vertices of the graph.

        Parameters:

            building_blocks:
                The building blocks which need to be mapped to
                vertices.

        Returns:

            Maps each building block in `building_blocks` to the
            :class:`.Vertex` instances it should be
            placed on.

        Raises:

            :class:`AssertionError`
                If the any building block does not have a
                valid number of functional groups.

            :class:`ValueError`
                If there are multiple building blocks with the same
                number of functional groups.

        """

        # Use tuple here because it prints nicely.
        allowed_degrees = tuple(cls._vertices_of_degree.keys())

        building_blocks_by_degree = {}
        for building_block in building_blocks:
            num_fgs = building_block.get_num_functional_groups()
            assert num_fgs in cls._vertices_of_degree.keys(), (
                "The number of functional groups in "
                f"{building_block} needs to be one of "
                f"{allowed_degrees}, but is "
                "currently "
                f"{building_block.get_num_functional_groups()}."
            )

            if num_fgs in building_blocks_by_degree:
                raise ValueError(
                    "If there are multiple building blocks with the "
                    "same number of functional groups, "
                    "building_block_vertices must be set explicitly."
                )
            building_blocks_by_degree[num_fgs] = building_block

        building_block_vertices: dict[BuildingBlock, list[_CageVertex]]
        building_block_vertices = {}
        for vertex in cls._vertex_prototypes:
            vertex_degree = cls._vertex_degrees[vertex.get_id()]
            building_block = building_blocks_by_degree[vertex_degree]
            building_block_vertices[
                building_block
            ] = building_block_vertices.get(building_block, [])
            building_block_vertices[building_block].append(vertex)
        return typing.cast(
            dict[BuildingBlock, abc.Sequence[_CageVertex]],
            building_block_vertices,
        )

    def _get_scale(
        self,
        building_block_vertices: dict[BuildingBlock, abc.Sequence[Vertex]],
    ) -> float:
        return max(bb.get_maximum_diameter() for bb in building_block_vertices)

    def _get_construction_state(self) -> _CageConstructionState:
        return _CageConstructionState(
            building_block_vertices=self._building_block_vertices,
            edges=self._edges,
            num_placement_stages=self._implementation.get_num_stages(),
            vertex_degrees=self._vertex_degrees,
        )

    def __repr__(self) -> str:
        vertex_alignments = (
            f"vertex_alignments={self._vertex_alignments}"
            if self._vertex_alignments
            else ""
        )
        return f"cage.{self.__class__.__name__}({vertex_alignments})"

    def with_building_blocks(
        self,
        building_block_map: dict[BuildingBlock, BuildingBlock],
    ) -> Cage:
        return self.clone()._with_building_blocks(building_block_map)



from collections.abc import Callable

from stk._internal.molecule import Molecule


class MoleculeKeyMaker:
    """
    An abstract base class for making :class:`.Molecule` keys.

    Keys are used in :mod:`stk` to determine if two molecules are
    duplicates of each other.

    Notes:

        You might notice that the public methods of this abstract base
        class are implemented. This is purely for convenience when
        implementing subclasses. The implemented public methods are
        simply default implementations, which can be safely ignored or
        overridden, when implementing subclasses.

    Examples:

        *Creating a New Key Maker Directly*

        Apart from using the subclasses provided, a
        :class:`.MoleculeKeyMaker` can be used directly , if you don't
        feel like writing a subclass

        .. testcode:: creating-a-new-key-maker-directly

            import stk

            # Create a MoleculeKeyMaker instance with a custom get_key
            # method.
            get_num_atoms = stk.MoleculeKeyMaker(
                key_name='num_atoms',
                get_key=lambda molecule: str(molecule.get_num_atoms()),
            )

            # Use the MoleculeKeyMaker instance.
            jsonizer = stk.MoleculeJsonizer(
                key_makers=(get_num_atoms, ),
            )
            # Get the JSON representation of a molecule.
            json = jsonizer.to_json(stk.BuildingBlock('NCCN'))

        .. testcode:: creating-a-new-key-maker-directly
            :hide:

            assert json['molecule']['num_atoms'] == '12'
            assert json['matrix']['num_atoms'] == '12'

    """

    def __init__(
        self,
        key_name: str,
        get_key: Callable[[Molecule], str],
    ) -> None:
        """
        Parameters:
            key_name:
                The name of the key.
            get_key:
                Takes a single parameter, `molecule`, and returns the
                key to use for that molecule. The value passed to the
                parameter must be a :class:`.Molecule` instance.
        """
        self._key_name = key_name
        self._get_key = get_key

    def get_key_name(self) -> str:
        """
        Get the name of the key.

        Returns:
            The name of the key.
        """
        return self._key_name

    def get_key(self, molecule: Molecule) -> str:
        """
        Get the key of `molecule`.

        Parameters:
            molecule:
                The molecule for which a key is needed.
        Returns:
            The key of `molecule`.
        """
        return self._get_key(molecule)

    def __str__(self) -> str:
        return repr(self)

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}("
            f"{self._key_name!r}, {self._get_key!r}"
            ")"
        )



import os

import numpy as np
import pytest
import stk
from pytest_lazyfixture import lazy_fixture
from rdkit.Chem import AllChem as rdkit

from .case_data import CaseData

# Fixtures need to be visible for lazy_fixture() calls.
from .fixtures import *  # noqa


@pytest.fixture(
    scope="session",
    params=[
        lambda: stk.BuildingBlock.init(
            atoms=(stk.C(0, 4),),
            bonds=(),
            position_matrix=np.array([[0.0, 0.0, 0.0]]),
            functional_groups=(),
        ),
        lambda: stk.BuildingBlock.init(
            atoms=(stk.C(0, 3), stk.H(1)),
            bonds=(stk.Bond(stk.C(0, 3), stk.H(1), 1),),
            position_matrix=np.array(
                [
                    [0.39382080513175644, 0.0, 0.0],
                    [-0.39382080513175644, 0.0, 0.0],
                ]
            ),
            functional_groups=(),
        ),
        lambda: stk.BuildingBlock.init(
            atoms=(stk.C(0, 2), stk.H(1), stk.H(2)),
            bonds=(
                stk.Bond(stk.C(0, 2), stk.H(1), 1),
                stk.Bond(stk.C(0, 2), stk.H(2), 1),
            ),
            position_matrix=np.array(
                [
                    [
                        -0.002271396061231665,
                        0.034037398527897535,
                        -0.0,
                    ],
                    [-1.0494595365731274, -0.017073891221884126, -0.0],
                    [1.0517309326343591, -0.016963507306017023, 0.0],
                ]
            ),
            functional_groups=(),
        ),
        lambda: stk.BuildingBlock("NCCN"),
        lambda: stk.BuildingBlock(
            "C(#Cc1cccc2ccncc21)c1ccc2[nH]c3ccc" "(C#Cc4cccc5cnccc54)cc3c2c1"
        ),
        lambda: stk.BuildingBlock(
            "C(#Cc1cccc2cnccc12)c1ccc2[nH]c3ccc" "(C#Cc4cccc5ccncc45)cc3c2c1"
        ),
        lambda: stk.BuildingBlock("N[C+][C+2]N"),
        lambda: stk.BuildingBlock("NCCN", [stk.PrimaryAminoFactory()]),
        lambda: stk.ConstructedMolecule(
            topology_graph=stk.polymer.Linear(
                building_blocks=(
                    stk.BuildingBlock("BrCCBr", [stk.BromoFactory()]),
                    stk.BuildingBlock(
                        smiles="BrCNCCBr",
                        functional_groups=[stk.BromoFactory()],
                    ),
                ),
                repeating_unit="AB",
                num_repeating_units=2,
            ),
        ),
    ],
)
def molecule(request) -> stk.Molecule:
    """
    A :class:`.Molecule` instance.

    """

    return request.param()


def get_random_position_matrix(molecule):
    generator = np.random.RandomState(4)
    return generator.normal(
        loc=23.3,
        scale=32.1,
        size=(molecule.get_num_atoms(), 3),
    )


@pytest.fixture(
    params=(
        lambda molecule: np.zeros((molecule.get_num_atoms(), 3)),
        lambda molecule: np.array(
            [[i, -i, 10.12 * i] for i in range(molecule.get_num_atoms())]
        ),
        lambda molecule: molecule.get_position_matrix(),
        get_random_position_matrix,
    ),
)
def get_position_matrix(request):
    """
    Return a valid position matrix for a molecule.

    Parameters
    ----------
    molecule : :class:`.Molecule`
        The molecule for which a position matrix is returned.

    Returns
    -------
    :class:`numpy.ndarray`
        A position matrix for `molecule`.

    """

    return request.param


@pytest.fixture(
    params=[
        [0, 0, 0],
        [10, 20, 30],
        [-10, 20, -30],
        [0.5, 10, -0.921],
    ]
)
def origin(request):
    return np.array(request.param)


@pytest.fixture
def get_origin_0(origin):
    return lambda molecule: origin


@pytest.fixture(
    params=[
        lambda molecule: molecule.get_centroid(),
        lazy_fixture("get_origin_0"),
    ],
)
def get_origin(request):
    """
    Return an origin parameter for a molecule.

    Parameters
    ----------
    molecule : :class:`.Molecule`
        The molecule which needs an origin parameter for some method.

    Returns
    -------
    :class:`numpy.ndarray`
        The origin parameter to use.

    """

    return request.param


@pytest.fixture(
    params=[
        lambda molecule: None,
        lambda molecule: 0,
        lambda molecule: range(molecule.get_num_atoms()),
        lambda molecule: range(0, molecule.get_num_atoms(), 2),
        lambda molecule: list(range(0, min(1, molecule.get_num_atoms()))),
        lambda molecule: tuple(range(0, min(1, molecule.get_num_atoms()))),
        lambda molecule: (
            i for i in range(0, min(1, molecule.get_num_atoms()))
        ),
        lambda molecule: (),
        lambda molecule: range(min(molecule.get_num_atoms(), 1)),
        lambda molecule: range(min(molecule.get_num_atoms(), 2)),
        lambda molecule: range(min(molecule.get_num_atoms(), 3)),
    ],
)
def get_atom_ids(request):
    """
    Return an atom_ids parameter for a :class:`.Molecule`.

    Parameters
    ----------
    molecule : :class:`.Molecule`
        The molecule for which `atom_ids` are returned.

    Retruns
    -------
    :class:`iterable` of :class:`int`
        An `atom_ids` parameter.

    """

    return request.param


def _get_dative_molecule() -> rdkit.Mol:
    dative_molecule = rdkit.MolFromSmiles("[Fe+2]<-N")
    dative_molecule.AddConformer(
        conf=rdkit.Conformer(dative_molecule.GetNumAtoms()),
    )
    return dative_molecule


@pytest.fixture(
    params=(
        lambda name: CaseData(
            molecule=stk.BuildingBlock("NCCN"),
            smiles="[H]N([H])C([H])([H])C([H])([H])N([H])[H]",
            name=name,
        ),
        lambda name: CaseData(
            molecule=stk.BuildingBlock("[H]NCCN"),
            smiles="[H]N([H])C([H])([H])C([H])([H])N([H])[H]",
            name=name,
        ),
        lambda name: CaseData(
            molecule=stk.BuildingBlock("C(N)CN"),
            smiles="[H]N([H])C([H])([H])C([H])([H])N([H])[H]",
            name=name,
        ),
        lambda name: CaseData(
            molecule=(
                stk.BuildingBlock("C(N)CN").with_canonical_atom_ordering()
            ),
            smiles="[H]N([H])C([H])([H])C([H])([H])N([H])[H]",
            name=name,
        ),
        lambda name: CaseData(
            molecule=(
                stk.BuildingBlock.init_from_rdkit_mol(
                    molecule=_get_dative_molecule(),
                )
            ),
            smiles="N->[Fe+2]",
            name=name,
        ),
        lambda name: CaseData(
            molecule=stk.BuildingBlock(
                smiles="[Fe+2]",
                position_matrix=[[0, 0, 0]],
            ),
            smiles="[Fe+2]",
            name=name,
        ),
        lambda name: CaseData(
            molecule=stk.BuildingBlock(
                smiles="[Fe+2]",
                position_matrix=np.array([[0, 0, 0]]),
            ),
            smiles="[Fe+2]",
            name=name,
        ),
    ),
)
def building_block(request) -> CaseData:
    return request.param(
        f"{request.fixturename}{request.param_index}",
    )


@pytest.fixture(
    params=(
        lazy_fixture("building_block"),
        lazy_fixture("cage"),
        lazy_fixture("cof"),
        lazy_fixture("polymer"),
        lazy_fixture("host_guest"),
        lazy_fixture("macrocycle"),
        lazy_fixture("metal_complex"),
        lazy_fixture("rotaxane"),
    ),
)
def case_data(request):
    """
    A :class:`.CaseData` instance.

    Notes
    -----
    This fixture either needs to load a saved position matrix for each
    test case, or it needs to save the position matrix of the molecule
    in the test case.

    The purpose of loading a saved position matrix, is to ensure that
    the geometry of the molecule being tested has not changed
    unexpectedly.

    If the position matrix held in a file needs to be updated, the file
    holding the position matrix simply needs to be deleted. When the
    file is missing, this fixture will save the current geometry of a
    molecule into the file.

    """

    path = os.path.join(
        os.path.dirname(os.path.realpath(__file__)),
        "fixtures",
        "position_matrices",
        f"{request.param.name}.npy",
    )

    if not os.path.exists(path):
        np.save(
            file=path,
            arr=request.param.molecule.get_position_matrix(),
            allow_pickle=False,
        )

    return request.param.with_position_matrix(path)


@pytest.fixture(
    scope="session",
)
def molecule_db(mongo_client):
    """
    A :class:`.MoleculeDatabase` instance.

    """

    return stk.MoleculeMongoDb(
        mongo_client=mongo_client,
        database="_stk_pytest_database",
        jsonizer=stk.MoleculeJsonizer(
            key_makers=(stk.Smiles(),),
        ),
        indices=(stk.Smiles().get_key_name(),),
    )


@pytest.fixture(
    scope="session",
)
def name_db(mongo_client):
    """
    A :class:`.ValueDatabase` for holding the names of molecules.

    """

    return stk.ValueMongoDb(
        mongo_client=mongo_client,
        database="_stk_pytest_database",
        collection="name",
        key_makers=(stk.Smiles(),),
        indices=(stk.Smiles().get_key_name(),),
    )



import itertools as it


def is_equivalent_atom(atom1, atom2):
    assert atom1.get_id() == atom2.get_id()
    assert atom1.get_charge() == atom2.get_charge()
    assert atom1.get_atomic_number() == atom2.get_atomic_number()
    assert atom1.__class__ is atom2.__class__


def is_clone_sequence(atoms, clones):
    """
    Test if atoms are correctly cloned.

    """

    for atom, clone in it.zip_longest(atoms, clones):
        assert atom is clone


def is_clone_functional_group(functional_group1, functional_group2):
    assert functional_group1 is not functional_group2

    is_clone_sequence(
        functional_group1.get_atoms(),
        functional_group2.get_atoms(),
    )
    same_ids(
        ids1=functional_group1.get_atom_ids(),
        ids2=functional_group2.get_atom_ids(),
    )
    same_ids(
        ids1=functional_group1.get_placer_ids(),
        ids2=functional_group2.get_placer_ids(),
    )
    same_ids(
        ids1=functional_group1.get_core_atom_ids(),
        ids2=functional_group2.get_core_atom_ids(),
    )


def is_clone_generic_functional_group(
    functional_group1,
    functional_group2,
):
    is_clone_functional_group(functional_group1, functional_group2)

    is_clone_sequence(
        functional_group1.get_bonders(),
        functional_group2.get_bonders(),
    )
    same_ids(
        functional_group1.get_bonder_ids(),
        functional_group2.get_bonder_ids(),
    )

    is_clone_sequence(
        functional_group1.get_deleters(),
        functional_group2.get_deleters(),
    )
    same_ids(
        functional_group1.get_deleter_ids(),
        functional_group2.get_deleter_ids(),
    )


def is_equivalent_sequence(atoms1, atoms2):
    for atom1, atom2 in it.zip_longest(atoms1, atoms2):
        is_equivalent_atom(atom1, atom2)


def same_ids(ids1, ids2):
    for id1, id2 in it.zip_longest(ids1, ids2):
        assert id1 == id2


def is_equivalent_functional_group(
    functional_group1,
    functional_group2,
):
    is_equivalent_sequence(
        functional_group1.get_atoms(),
        functional_group2.get_atoms(),
    )
    same_ids(
        ids1=functional_group1.get_atom_ids(),
        ids2=functional_group2.get_atom_ids(),
    )
    same_ids(
        ids1=functional_group1.get_placer_ids(),
        ids2=functional_group2.get_placer_ids(),
    )
    same_ids(
        ids1=functional_group1.get_core_atom_ids(),
        ids2=functional_group2.get_core_atom_ids(),
    )


def is_equivalent_generic_functional_group(
    functional_group1,
    functional_group2,
):
    is_equivalent_functional_group(
        functional_group1=functional_group1,
        functional_group2=functional_group2,
    )
    is_equivalent_sequence(
        functional_group1.get_bonders(),
        functional_group2.get_bonders(),
    )
    same_ids(
        ids1=functional_group1.get_bonder_ids(),
        ids2=functional_group2.get_bonder_ids(),
    )
    assert (
        functional_group1.get_num_bonders()
        == functional_group2.get_num_bonders()
    )

    is_equivalent_sequence(
        functional_group1.get_deleters(),
        functional_group2.get_deleters(),
    )
    same_ids(
        ids1=functional_group1.get_deleter_ids(),
        ids2=functional_group2.get_deleter_ids(),
    )



def is_equivalent_atom(atom1, atom2):
    assert atom1.get_id() == atom2.get_id()
    assert atom1.get_charge() == atom2.get_charge()
    assert atom1.__class__ is atom2.__class__
    assert atom1.get_atomic_number() == atom2.get_atomic_number()



import numpy as np
import pytest
from pytest_lazyfixture import lazy_fixture

# Fixtures need to be visible for lazy_fixture() calls.
from .fixtures import *  # noqa


@pytest.fixture(
    params=(
        lazy_fixture("metal"),
        lazy_fixture("monodentate"),
        lazy_fixture("bidentate"),
    ),
)
def case_data(request):
    """
    A :class:`.CaseData` instance.

    """

    return request.param


@pytest.fixture(
    params=([1, 2, -20],),
)
def position(request):
    """
    The `position` of a vertex.

    """

    return np.array(request.param, dtype=np.float64)



import itertools as it

import numpy as np
import pytest
import stk

from ..case_data import CaseData


@pytest.fixture
def functional_group2_2(functional_group2):
    """
    A :class:`.GenericFunctionalGroup` instance with 2 bonder atoms.

    """

    return functional_group2


@pytest.fixture
def two_two_reaction(
    functional_group2,
    functional_group2_2,
    bond_order,
    periodicity,
):
    """
    A :class:`.CaseData` instance.

    """

    return CaseData(
        reaction=stk.TwoTwoReaction(
            construction_state=get_construction_state(
                functional_group1=functional_group2_2,
                functional_group2=functional_group2,
            ),
            functional_group1=functional_group2_2,
            functional_group2=functional_group2,
            bond_order=bond_order,
            periodicity=periodicity,
        ),
        new_atoms=(),
        new_bonds=tuple(
            get_bonds(
                functional_group1=functional_group2_2,
                functional_group2=functional_group2,
                bond_order=bond_order,
                periodicity=periodicity,
            ),
        ),
        deleted_atoms=tuple(
            it.chain(
                functional_group2_2.get_deleters(),
                functional_group2.get_deleters(),
            )
        ),
        deleted_bonds=(),
    )


def get_bonds(
    functional_group1,
    functional_group2,
    bond_order,
    periodicity,
):
    b1, b2 = functional_group1.get_bonders()
    b3, b4 = functional_group2.get_bonders()
    yield stk.Bond(
        atom1=b1,
        atom2=b3,
        order=bond_order,
        periodicity=periodicity,
    )

    yield stk.Bond(
        atom1=b2,
        atom2=b4,
        order=bond_order,
        periodicity=periodicity,
    )


class MockConstructionState(stk.ConstructionState):
    def __init__(self, position_matrix):
        self.position_matrix = position_matrix

    def get_position_matrix(self):
        return self.position_matrix


def get_construction_state(functional_group1, functional_group2):
    size = max(
        it.chain(
            functional_group1.get_bonder_ids(),
            functional_group2.get_bonder_ids(),
        )
    )
    position_matrix = np.zeros((size + 1, 3))
    b1, b2 = functional_group1.get_bonder_ids()
    b3, b4 = functional_group2.get_bonder_ids()
    position_matrix[b1] = [100, 100, 100]
    position_matrix[b3] = [99, 99, 99]
    position_matrix[b2] = [10, 10, 10]
    position_matrix[b4] = [8, 8, 8]
    return MockConstructionState(position_matrix)



import pytest
import stk

from ..case_data import GenericCaseData


@pytest.fixture
def dibromo(get_atom_ids):
    a, b, c, d = get_atom_ids(4)
    return _dibromo(stk.Br(a), stk.C(b), stk.Br(c), stk.C(d))


def _dibromo(bromine1, atom1, bromine2, atom2):
    bonders = (atom1, atom2)
    deleters = (bromine1, bromine2)
    return GenericCaseData(
        functional_group=stk.Dibromo(
            bromine1=bromine1,
            atom1=atom1,
            bromine2=bromine2,
            atom2=atom2,
            bonders=bonders,
            deleters=deleters,
        ),
        atoms=(bromine1, atom1, bromine2, atom2),
        bonders=bonders,
        deleters=deleters,
    )



import stk

from ..utilities import is_equivalent_functional_group


def test_repr(functional_group):
    """
    Test :meth:`.FunctionalGroup.__repr__`.

    Parameters
    ----------
    functional_group : :class:`.FunctionalGroup`
        The functional group whose representation is tested.

    Returns
    -------
    None : :class:`NoneType`

    """

    other = eval(repr(functional_group), dict(stk.__dict__))
    is_equivalent_functional_group(functional_group, other)



import pytest
import stk

from ...case_data import CaseData


@pytest.fixture(
    scope="session",
    params=(
        lambda name: CaseData(
            molecule=stk.ConstructedMolecule(
                topology_graph=stk.polymer.Linear(
                    building_blocks=(
                        stk.BuildingBlock(
                            smiles="BrC1=C(Br)[C+]=N1",
                            functional_groups=[stk.BromoFactory()],
                        ),
                        stk.BuildingBlock(
                            smiles="Br[C+]=NC#CBr",
                            functional_groups=[stk.BromoFactory()],
                        ),
                    ),
                    repeating_unit="AB",
                    num_repeating_units=2,
                ),
            ),
            smiles="BrC#CN=[C+]C1=C(C#CN=[C+]C2=C(Br)N=[C+]2)N=[C+]1",
            name=name,
        ),
        lambda name: CaseData(
            molecule=stk.ConstructedMolecule(
                topology_graph=stk.polymer.Linear(
                    building_blocks=(
                        stk.BuildingBlock(
                            smiles="BrC1=C(Br)[C+]=N1",
                            functional_groups=[stk.BromoFactory()],
                        ),
                        stk.BuildingBlock(
                            smiles="Br[C+]=NC#CBr",
                            functional_groups=[stk.BromoFactory()],
                        ),
                    ),
                    repeating_unit="AB",
                    num_repeating_units=2,
                    optimizer=stk.Collapser(scale_steps=False),
                ),
            ),
            smiles="BrC#CN=[C+]C1=C(C#CN=[C+]C2=C(Br)N=[C+]2)N=[C+]1",
            name=name,
        ),
        lambda name: CaseData(
            molecule=stk.ConstructedMolecule(
                topology_graph=stk.polymer.Linear(
                    building_blocks=(
                        stk.BuildingBlock(
                            smiles="BrC1=C(Br)N=[C+]1",
                            functional_groups=[stk.BromoFactory()],
                        ),
                    ),
                    repeating_unit="A",
                    num_repeating_units=1,
                ),
            ),
            smiles="BrC1=C(Br)N=[C+]1",
            name=name,
        ),
        lambda name: CaseData(
            molecule=stk.ConstructedMolecule(
                topology_graph=stk.polymer.Linear(
                    building_blocks=(
                        stk.BuildingBlock(
                            smiles="BrCBr",
                            functional_groups=[stk.BromoFactory()],
                        ),
                        stk.BuildingBlock(
                            smiles="BrCCCBr",
                            functional_groups=[stk.BromoFactory()],
                        ),
                    ),
                    repeating_unit="AB",
                    num_repeating_units=3,
                ),
            ),
            smiles=(
                "[H]C([H])(Br)C([H])([H])C([H])([H])C([H])([H])C([H])("
                "[H])C([H])([H])C([H])([H])C([H])([H])C([H])([H])C([H]"
                ")([H])C([H])([H])C([H])([H])Br"
            ),
            name=name,
        ),
    ),
)
def polymer_linear(request) -> CaseData:
    return request.param(
        f"{request.fixturename}{request.param_index}",
    )



"""
Square
======

"""

import numpy as np

from stk._internal.topology_graphs.edge import Edge

from .cof import Cof
from .vertices import LinearVertex, NonLinearVertex


class Square(Cof):
    """
    Represents a sqaure COF topology graph.

    Unoptimized construction

    .. moldoc::

        import moldoc.molecule as molecule
        import stk

        cof = stk.ConstructedMolecule(
            topology_graph=stk.cof.PeriodicSquare(
                building_blocks=(
                    stk.BuildingBlock(
                        smiles='BrCC(Br)',
                        functional_groups=[stk.BromoFactory()],
                    ),
                    stk.BuildingBlock(
                        smiles='BrC1=C(Br)C(Br)=C1Br',
                        functional_groups=[stk.BromoFactory()],
                    ),
                ),
                lattice_size=(3, 3, 1),
            ),
        )
        moldoc_display_molecule = molecule.Molecule(
            atoms=(
                molecule.Atom(
                    atomic_number=atom.get_atomic_number(),
                    position=position,
                ) for atom, position in zip(
                    cof.get_atoms(),
                    cof.get_position_matrix(),
                )
            ),
            bonds=(
                molecule.Bond(
                    atom1_id=bond.get_atom1().get_id(),
                    atom2_id=bond.get_atom2().get_id(),
                    order=bond.get_order(),
                ) for bond in cof.get_bonds()
                if all(p == 0 for p in bond.get_periodicity())
            ),
        )

    ``Collapser(scale_steps=False)`` optimized construction

    .. moldoc::

        import moldoc.molecule as molecule
        import stk

        cof = stk.ConstructedMolecule(
            topology_graph=stk.cof.PeriodicSquare(
                building_blocks=(
                    stk.BuildingBlock(
                        smiles='BrCC(Br)',
                        functional_groups=[stk.BromoFactory()],
                    ),
                    stk.BuildingBlock(
                        smiles='BrC1=C(Br)C(Br)=C1Br',
                        functional_groups=[stk.BromoFactory()],
                    ),
                ),
                lattice_size=(3, 3, 1),
                optimizer=stk.Collapser(scale_steps=False),
            ),
        )
        moldoc_display_molecule = molecule.Molecule(
            atoms=(
                molecule.Atom(
                    atomic_number=atom.get_atomic_number(),
                    position=position,
                ) for atom, position in zip(
                    cof.get_atoms(),
                    cof.get_position_matrix(),
                )
            ),
            bonds=(
                molecule.Bond(
                    atom1_id=bond.get_atom1().get_id(),
                    atom2_id=bond.get_atom2().get_id(),
                    order=bond.get_order(),
                ) for bond in cof.get_bonds()
                if all(p == 0 for p in bond.get_periodicity())
            ),
        )

    Building blocks with four and two functional groups are required
    for this topology graph.

    When using a :class:`dict` for the `building_blocks` parameter,
    as in :ref:`cof-topology-graph-examples`:
    *Multi-Building Block COF Construction*, a
    :class:`.BuildingBlock`, with the following number of functional
    groups, needs to be assigned to each of the following vertex ids:

        | 4-functional groups: 0
        | 2-functional groups: 1 to 2

    See :class:`.Cof` for more details and examples.

    """

    _lattice_constants = _a, _b, _c = (
        np.array([1.0, 0.0, 0.0]),
        np.array([0.0, 1.0, 0.0]),
        np.array([0.0, 0.0, 1.0]),
    )

    _non_linears = (NonLinearVertex(0, (0.5) * _a + (0.5) * _b + (0.5) * _c),)
    _vertex_prototypes = (
        *_non_linears,
        LinearVertex.init_at_shifted_center(
            id=1,
            vertices=(_non_linears[0], _non_linears[0]),
            cell_shifts=((0, 0, 0), (1, 0, 0)),
            lattice_constants=_lattice_constants,
        ),
        LinearVertex.init_at_shifted_center(
            id=2,
            vertices=(_non_linears[0], _non_linears[0]),
            cell_shifts=((0, 0, 0), (0, 1, 0)),
            lattice_constants=_lattice_constants,
        ),
    )

    _edge_prototypes = (
        Edge(0, _vertex_prototypes[1], _vertex_prototypes[0]),
        Edge(
            id=1,
            vertex1=_vertex_prototypes[1],
            vertex2=_vertex_prototypes[0],
            periodicity=(1, 0, 0),
        ),
        Edge(2, _vertex_prototypes[2], _vertex_prototypes[0]),
        Edge(
            id=3,
            vertex1=_vertex_prototypes[2],
            vertex2=_vertex_prototypes[0],
            periodicity=(0, 1, 0),
        ),
    )



import numpy as np
import rdkit.Chem.AllChem as rdkit

from ...utilities import is_clone


def test_with_canonical_atom_ordering(case_data):
    """
    Test :meth:`.Molecule.with_canonical_atom_ordering`.

    Parameters
    ----------
    case_data : :class:`.CaseData`
        A test case. Holds the molecule to test and the expected
        result.

    Returns
    -------
    None : :class:`NoneType`

    """

    _test_with_canonical_atom_ordering(
        molecule=case_data.molecule,
        result=case_data.result,
    )


def _test_with_canonical_atom_ordering(molecule, result):
    """
    Test :meth:`.Molecule.with_canonical_atom_ordering`.

    Parameters
    ----------
    molecule : :class:`.Molecule`
        The molecule to test.

    result : :class:`.Molecule`
        The molecule with canonical ordering.

    Returns
    -------
    None : :class:`NoneType`

    """

    ordered = molecule.with_canonical_atom_ordering()
    is_clone(ordered, result)
    order = rdkit.CanonicalRankAtoms(molecule.to_rdkit_mol())
    old_position_matrix = molecule.get_position_matrix()
    new_position_matrix = ordered.get_position_matrix()
    for old_id, new_id in enumerate(order):
        assert np.all(
            np.equal(
                old_position_matrix[old_id],
                new_position_matrix[new_id],
            )
        )



def test_write(case_data, tmp_path):
    """
    Test writing of molecule to a file.

    Parameters
    ----------
    case_data : :class:`.CaseData`
        A test case.

    tmp_path : :class:`pathlib2.Path`
        Path to temporary directory.

    Returns
    -------
    None : :class:`NoneType`

    """

    _test_write(
        molecule=case_data.molecule,
        writer=case_data.writer,
        string=case_data.string,
        file_path=tmp_path / "tmp.mol",
    )


def _test_write(molecule, writer, string, file_path):
    """
    Test that the written file content matches expected string.

    Parameters
    ----------
    molecule : :class:`.Molecule`
        Molecule to test.

    writer : :class:`.MolWriter`
        The writer to test.

    string : :class:`str`
        The expected output string.

    file_path : :class:`str`
        Path to temporary file.

    Returns
    -------
    None : :class:`NoneType`

    """

    writer.write(molecule, file_path)

    with open(file_path, "r") as f:
        content = f.read()

    assert content == string



import numpy as np
import pytest


class ScaleData:
    """
    Parameters for a :meth:`.Vertex.with_scale` test.

    Attributes
    ----------
    start : :class:`numpy.ndarray`
        The starting position of the tested vertex.

    scale : :class:`float` or :class:`numpy.ndarray`
        The scale applied to tested vertex.

    target : :class:`numpy.ndarray`
        The position of the new vertex.

    """

    def __init__(self, start, scale, target):
        """
        Initialize a :class:`.ScaleData` instance.

        Parameters
        ----------
        start : :class:`numpy.ndarray`
            The starting position of the tested vertex.

        scale : :class:`float` or :class:`numpy.ndarray`
            The scale applied to tested vertex.

        target : :class:`numpy.ndarray`
            The position of the new vertex.

        """

        self.start = start
        self.scale = scale
        self.target = target


@pytest.fixture(
    params=(
        ScaleData(np.array([0, 0, 0]), 2, np.array([0, 0, 0])),
        ScaleData(np.array([0, 0, 0]), 0, np.array([0, 0, 0])),
        ScaleData(np.array([1, -1, 0]), 0, np.array([0, 0, 0])),
        ScaleData(np.array([1, -1, 2]), 2, np.array([2, -2, 4])),
        ScaleData(np.array([2, 4, -5]), -3, np.array([-6, -12, 15])),
        ScaleData(
            np.array([1, -1, 2]),
            np.array([10, 2, -2]),
            np.array([10, -2, -4]),
        ),
    ),
)
def scale_data(request):
    """
    A :class:`.ScaleData` instance.

    """

    return request.param


@pytest.fixture(
    params=(
        [0, 0, 0],
        [1, 2, -20],
    ),
)
def position(request):
    return np.array(request.param, dtype=np.float64)


@pytest.fixture
def vertex(case_data):
    return case_data.vertex



"""
Edge Sorter
===========

"""

from .sorter import _Sorter


class _EdgeSorter(_Sorter):
    """
    Sorted edges according to their angle.

    """

    __slots__ = [
        "_items",
        "_reference",
        "_axis",
        "_edge_centroid",
    ]

    def __init__(self, edges, aligner_edge, axis):
        """
        Initialize an :class:`._EdgeSorter` instance.

        Parameters
        ----------
        edges : :class:`iterable` of :class:`.Edge`
            The edges to sort.

        aligner_edge : :class:`.Edge`
            The edge in edges, used to calculate the reference vector.

        axis : :class:`numpy.ndarray`
            Must be immutable. The axis used to determine the clockwise
            direction.

        """

        self._edge_centroid = edge_centroid = sum(
            edge.get_position() for edge in edges
        ) / len(edges)
        super().__init__(
            items=edges,
            axis=axis,
            reference=aligner_edge.get_position() - edge_centroid,
        )

    def _get_vector(self, item):
        return item.get_position() - self._edge_centroid



import stk

from .case_data import CaseData


def test_get_num_placers(case_data: CaseData) -> None:
    """
    Test :meth:`.BuildingBlock.get_num_placers`.

    Parameters:

        case_data:
            A test case. Holds the building block to test and the
            correct number of *placer* atoms.

    """

    _test_get_num_placers(
        building_block=case_data.building_block,
        expected_num_placers=len(case_data.placer_ids),
    )


def _test_get_num_placers(
    building_block: stk.BuildingBlock,
    expected_num_placers: int,
) -> None:
    """
    Test :meth:`.BuildingBlock.get_num_placers`.

    Parameters:

        building_block:
            The building block to test.

        expected_num_placers:
            The correct number of *placer* atoms

    """

    assert building_block.get_num_placers() == expected_num_placers



from collections.abc import Callable, Iterable
from typing import Any

import numpy as np

from stk._internal.ea.molecule_record import MoleculeRecord
from stk._internal.ea.mutation.record import MutationRecord
from stk._internal.topology_graphs.topology_graph.topology_graph import (
    TopologyGraph,
)


class RandomTopologyGraph:
    """
    Changes topology graphs at random.

    Examples:
        *Constructed Molecule Mutation*

        .. testcode:: constructed-molecule-mutation

            import stk

            # Create a molecule which is to be mutated.
            bb1 = stk.BuildingBlock('NCCN', [stk.PrimaryAminoFactory()])
            bb2 = stk.BuildingBlock(
                smiles='O=CCC(C=O)CC=O',
                functional_groups=[stk.AldehydeFactory()],
            )
            cage = stk.MoleculeRecord(
                topology_graph=stk.cage.FourPlusSix((bb1, bb2)),
            )

            # Create functions which replace the topology graph.
            replacement_funcs = (
                lambda graph:
                    stk.cage.TwoPlusThree(graph.get_building_blocks()),
                lambda graph:
                    stk.cage.EightPlusTwelve(graph.get_building_blocks()),
                lambda graph:
                    stk.cage.TwentyPlusThirty(graph.get_building_blocks()),
            )

            # Create the mutator.
            random_topology = stk.RandomTopologyGraph(replacement_funcs)

            # Mutate a molecule.
            mutation_record1 = random_topology.mutate(cage)

            # Mutate the molecule a second time.
            mutation_record2 = random_topology.mutate(cage)

    """

    def __init__(
        self,
        replacement_funcs: Iterable[Callable[[TopologyGraph], TopologyGraph]],
        name: str = "RandomTopologyGraph",
        random_seed: int | np.random.Generator | None = None,
    ) -> None:
        """
        Parameters:
            replacement_funcs \
(list[collections.abc.Callable[[TopologyGraph], TopologyGraph]]):
                Functions which take a topology graph and return
                its replacement. One is selected at random each
                time :meth:`.mutate` is called.

            name:
                A name to help identify the mutator instance.

            random_seed:
                The random seed to use.

        """

        if random_seed is None or isinstance(random_seed, int):
            random_seed = np.random.default_rng(random_seed)

        self._replacement_funcs = tuple(replacement_funcs)
        self._name = name
        self._generator = random_seed

    def mutate(
        self,
        record: MoleculeRecord[Any],
    ) -> MutationRecord[MoleculeRecord[Any]]:
        """
        Return a mutant of `record`.

        Parameters:
            record:
                The molecule to be mutated.

        Returns:
            A record of the mutation or ``None``
            if `record` cannot be mutated.
        """
        replacement_func = self._generator.choice(
            a=self._replacement_funcs,  # type: ignore
        )
        replacement = replacement_func(record.get_topology_graph())
        return MutationRecord(
            molecule_record=MoleculeRecord(replacement),
            mutator_name=self._name,
        )



import numpy as np


def test_get_a(periodic_case):
    """
    Test :meth:`.PeriodicInfo.get_a`.

    Parameters
    ----------
    periodic_case : :class:`.CaseData`
        The periodic case to test.

    Returns
    -------
    None : :class:`NoneType`

    """

    assert np.isclose(
        periodic_case.a,
        periodic_case.periodic_info.get_a(),
        atol=1e-6,
    )



import numpy as np

from ...utilities import get_num_atom_ids, normalize_ids
from ..utilities import get_direction


def test_get_direction(case_data, get_atom_ids):
    """
    Test :meth:`.Molecule.get_direction`.

    Parameters
    ----------
    case_data : :class:`.CaseData`
        A test case. Holds the molecule to test and its correct
        atomic positions.

    get_atom_ids : :class:`callable`
        Takes a single parameter, `molecule`, and returns a valid
        `atom_ids` parameter for :meth:`.Molecule.get_direction`. This
        allows the testing of different values of this parameter.

    Returns
    -------
    None : :class:`NoneType`

    Notes
    -----
    This test compares the result of :meth:`.Molecule.get_direction`
    to the result of :func:`.get_direction`, which is a utility
    function defined for the purposes of this test. Because
    :func:`.get_direction` is tested independently, in
    :mod:`.test_get_direction_helper`, if its tests pass, then
    it can be assumed, that :func:`.get_direction` gives correct
    results.

    Now, assuming that :func:`.get_direction` passed all of its tests,
    this test compares the results of :meth:`.Molecule.get_direction`
    to the results of :func:`.get_direction`. If the results do not
    match, the fault can be placed on :meth:`.Molecule.get_direction`,
    because :func:`.get_direction` has already been verified to be
    correct by its own tests.

    """

    _test_get_direction(
        molecule=case_data.molecule,
        direction=get_direction(
            position_matrix=case_data.position_matrix,
            atom_ids=tuple(
                normalize_ids(
                    molecule=case_data.molecule,
                    ids=get_atom_ids(case_data.molecule),
                )
            ),
        ),
        get_atom_ids=get_atom_ids,
    )


def _test_get_direction(molecule, direction, get_atom_ids):
    """
    Test :meth:`.Molecule.get_direction`.

    Parameters
    ----------
    molecule : :class:`.Molecule`
        The molecule to test.

    direction : :class:`.Molecule`
        The correct direction of `molecule`.

    get_atom_ids : :class:`callable`
        Takes a single parameter, `molecule`, and returns a valid
        `atom_ids` parameter for :meth:`.Molecule.get_direction`. This
        allows the testing of different values of this parameter.

    Returns
    -------
    None : :class:`NoneType`

    """

    if get_num_atom_ids(molecule, get_atom_ids) == 1:
        # Any non-0 vector is valid in this case.
        assert not np.allclose(
            a=[0, 0, 0],
            b=molecule.get_direction(get_atom_ids(molecule)),
            atol=1e-13,
        )
        return

    result = molecule.get_direction(get_atom_ids(molecule))
    # The direction may be parallel or anti-parallel.
    return abs(abs(result @ direction) - 1) < 1e-13
